# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015, Ignacio Fdez. Galván                             *
#***********************************************************************

from docutils.transforms import Transform
from docutils import nodes

# Change the "program" role to uppercase
#
class UppercaseProgram(Transform):
  default_priority = 10

  def apply(self):
    for node in self.document.traverse(nodes.inline):
      if ('program' in node['classes']):
        for child in node.traverse(nodes.Text):
          node.replace(child, nodes.Text(child.astext().upper()))

# Setup
#
def setup(app):
  app.add_transform(UppercaseProgram)



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

from typing import Sequence
from os import PathLike


from analyze_molcas._base_orbfile import (
    _Orbitals, FileFormat, _reindex, _spin_reindex,
    _read_spin_orbs, _read_header,
    _write_section_1D, _write_section_2D)


class SpinOrbs(_Orbitals):

    @classmethod
    def read_orbfile(cls, path: PathLike):
        with open(path, 'r') as f:
            version, orbs, spin_orbs = _read_header(f)

            if not spin_orbs:
                raise FileFormat('File represents spatial orbitals.')

            coeff, occ, energy, idx = _read_spin_orbs(f, orbs, version)
        return cls(orbs, coeff, occ, energy, idx)

    def reindex(self,
                new_idx: Sequence[Sequence[int]], inplace: bool=False):
        if inplace:
            self.coeff = {
                spin:
                    [coeff[:, idx] for idx, coeff in zip(new_idx, spin_values)]
                for spin, spin_values in self.coeff.items()}
            self.energy = _spin_reindex(self.energy, new_idx)
            self.idx = _reindex(self.idx, new_idx)
            self.occ = _spin_reindex(self.occ, new_idx)
        else:
            new = self.copy()
            new.reindex(new_idx, inplace=True)
            return new

    def _write_header(self):
        return ['#INPORB 2.2',
                '#INFO',
                '* PyOrb written',
                f"{1:8d}{len(self.orbs):8d}{0:8d}",
                f"{''.join(f'{x:8d}' for x in self.orbs)}",
                f"{''.join(f'{x:8d}' for x in self.orbs)}",
                '* SpinOrbs written from analyze_molcas']

    def _write_MO_coeff(self):
        def get_paragraph(irrep, orb):
            return f'* ORBITAL{irrep + 1:5d}{orb + 1:5d}'

        yield from _write_section_2D(
            title='#ORB', paragraph=get_paragraph,
            orbs=self.orbs, values=self.coeff['a'])
        yield from _write_section_2D(
            title='#UORB', paragraph=get_paragraph,
            orbs=self.orbs, values=self.coeff['b'])

    def _write_MO_occ(self):
        yield from _write_section_1D(
            title='#OCC', subtitle='* OCCUPATION NUMBERS',
            orbs=self.orbs, values=self.occ['a'])
        yield from _write_section_1D(
            title='#UOCC', subtitle='* Beta OCCUPATION NUMBERS',
            orbs=self.orbs, values=self.occ['b'])

    def _write_MO_human_occ(self):
        yield from _write_section_1D(
            title='#OCHR', subtitle='* OCCUPATION NUMBERS (HUMAN-READABLE)',
            orbs=self.orbs, values=self.occ['a'], cols=10, fmt='8.4f')
        yield from _write_section_1D(
            title='#UOCHR',
            subtitle='* Beta OCCUPATION NUMBERS (HUMAN-READABLE)',
            orbs=self.orbs, values=self.occ['b'], cols=10, fmt='8.4f')

    def _write_MO_E(self):
        yield from _write_section_1D(
            title='#ONE', subtitle='* ONE ELECTRON ENERGIES',
            orbs=self.orbs, values=self.energy['a'], cols=10, fmt='12.4E')
        yield from _write_section_1D(
            title='#UONE', subtitle='* Beta ONE ELECTRON ENERGIES',
            orbs=self.orbs, values=self.energy['b'], cols=10, fmt='12.4E')



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015,2016, Ignacio Fdez. Galván                        *
#***********************************************************************

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from docutils.statemachine import ViewList
from sphinx.util.nodes import set_source_info

class float_container(nodes.General, nodes.Element): pass
class table_container(nodes.General, nodes.Element): pass
class figure_container(nodes.General, nodes.Element): pass
class code_container(nodes.General, nodes.Element): pass

# Create the float directive
#
class FloatDirective(Directive):
  has_content = True
  required_arguments = 0
  optional_arguments = 0
  final_argument_whitespace = False
  option_spec = {
    'type': directives.unchanged_required,
    'caption': directives.unchanged_required,
    'caption-top': directives.flag,
    'name': directives.unchanged,
  }

  def run(self):
    if (self.options.get('type') == 'table'):
      content_node = table_container('')
    elif (self.options.get('type') == 'figure'):
      content_node = figure_container('')
    elif (self.options.get('type') == 'code'):
      content_node = code_container('')
    else:
      content_node = float_container('')
    self.state.nested_parse(self.content, self.content_offset, content_node)
    set_source_info(self, content_node)
    self.add_name(content_node)
    caption = self.options.get('caption')
    if caption:
      parsed = nodes.Element()
      self.state.nested_parse(ViewList([caption], source=''), self.content_offset, parsed)
      caption_node = nodes.caption(parsed[0].rawsource, '', *parsed[0].children)
      caption_node.source = parsed[0].source
      caption_node.line = parsed[0].line
      if ('caption-top' in self.options):
        content_node.insert(0, caption_node)
      else:
        content_node += caption_node
    return [content_node]

def visit_float_html(self, node):
  self.body.append(self.starttag(node, 'div', CLASS='float-wrapper'))

def depart_float_html(self, node):
  self.body.append('</div>\n')

def visit_float_latex(self, node):
  ids = ''
  if node['ids']:
    ids += self.hypertarget(node['ids'][0], anchor=False)
  if (isinstance(node, table_container)):
    floattype = 'table'
  elif (isinstance(node, figure_container)):
    floattype = 'figure'
  elif (isinstance(node, code_container)):
    floattype = 'code'
  else:
    floattype = 'float'
  self.body.append('\\begin{%s}\n' % floattype)
  if any(isinstance(child, nodes.caption) for child in node):
    self.body.append('\\capstart\n')
  self.context.append(ids + '\\end{%s}\n' % floattype)

def depart_float_latex(self, node):
  self.body.append(self.context.pop())

# Setup
#
def setup(app):
  app.add_directive('float', FloatDirective)
  app.add_node(float_container,
               html=(visit_float_html, depart_float_html),
               latex=(visit_float_latex, depart_float_latex))
  app.add_enumerable_node(table_container, 'table',
               html=(visit_float_html, depart_float_html),
               latex=(visit_float_latex, depart_float_latex))
  app.add_enumerable_node(figure_container, 'figure',
               html=(visit_float_html, depart_float_html),
               latex=(visit_float_latex, depart_float_latex))
  app.add_enumerable_node(code_container, 'code-block',
               html=(visit_float_html, depart_float_html),
               latex=(visit_float_latex, depart_float_latex))



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='analyze_molcas',
    version='0.0.0',
    license='LGPLv2',
    description='Module for molcas utility functions.',
    author='Oskar Weser',
    author_email='oskar.weser@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Utilities',
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires='>=3.6',
    install_requires=[
        'click', 'attrs'
    ],
    #  extras_require={
    #      # eg:
    #      #   'rst': ['docutils>=0.11'],
    #      #   ':python_version=="2.6"': ['argparse'],
    #  },
    entry_points={
        'console_scripts': [
            'spat_to_spin = analyze_molcas.cli:spat_to_spin',
            'assimilate = analyze_molcas.cli:assimilate',
            'combine_orbs = analyze_molcas.cli:combine_orbs',
        ]
    },
)



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015, Ignacio Fdez. Galván                             *
#***********************************************************************

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx.util.nodes import set_source_info
from sphinx.directives.code import container_wrapper, CodeBlock
import codecs
import os, os.path

# Create the extractfile directive
#
class ExtractFileDirective(Directive):
  has_content = True
  required_arguments = 1
  optional_arguments = 0
  final_argument_whitespace = False
  option_spec = {
  }

  def run(self):
    code = u'\n'.join(self.content)
    literal = nodes.literal_block(code, code)
    filename = self.arguments[0]
    addCode(self.state.document.settings.env, filename, code)
    return [literal]

# This is just a copy of the CodeBlock directive,
# with the filename argument and the additional extracting code
#
class ExtractCodeBlock(CodeBlock):
  has_content = True
  required_arguments = 0
  optional_arguments = 1
  final_argument_whitespace = False
  option_spec = {
    'force': directives.flag,
    'linenos': directives.flag,
    'dedent': int,
    'lineno-start': int,
    'emphasize-lines': directives.unchanged_required,
    'caption': directives.unchanged_required,
    'class': directives.class_option,
    'name': directives.unchanged,
    ###########################################
    'filename': directives.unchanged_required,
    ###########################################
  }
  def run(self):
    document = self.state.document
    code = '\n'.join(self.content)
    location = self.state_machine.get_source_and_line(self.lineno)
    linespec = self.options.get('emphasize-lines')
    if linespec:
      try:
        nlines = len(self.content)
        hl_lines = parselinenos(linespec, nlines)
        if any(i >= nlines for i in hl_lines):
          logger.warning(__('line number spec is out of range(1-%d): %r') %
                         (nlines, self.options['emphasize-lines']),
                         location=location)
        hl_lines = [x + 1 for x in hl_lines if x < nlines]
      except ValueError as err:
        return [document.reporter.warning(err, line=self.lineno)]
    else:
      hl_lines = None
    if 'dedent' in self.options:
      location = self.state_machine.get_source_and_line(self.lineno)
      lines = code.split('\n')
      lines = dedent_lines(lines, self.options['dedent'], location=location)
      code = '\n'.join(lines)
    literal = nodes.literal_block(code, code)  # type: Element
    if 'linenos' in self.options or 'lineno-start' in self.options:
      literal['linenos'] = True
    literal['classes'] += self.options.get('class', [])
    literal['force'] = 'force' in self.options
    if self.arguments:
      literal['language'] = self.arguments[0]
    else:
      literal['language'] = self.env.temp_data.get('highlight_language',
                                                   self.config.highlight_language)
    extra_args = literal['highlight_args'] = {}
    if hl_lines is not None:
      extra_args['hl_lines'] = hl_lines
    if 'lineno-start' in self.options:
      extra_args['linenostart'] = self.options['lineno-start']
    self.set_source_info(literal)
    caption = self.options.get('caption')
    if caption:
      try:
        literal = container_wrapper(self, literal, caption)
      except ValueError as exc:
        return [document.reporter.warning(exc, line=self.lineno)]
    self.add_name(literal)
    ###########################################
    filename = self.options.get('filename')
    addCode(self.state.document.settings.env, filename, code)
    ###########################################
    return [literal]
  def set_source_info(self, literal):
    try:
      super(ExtractCodeBlock, self).set_source_info(literal)
    except AttributeError:
      set_source_info(self, literal)

# Add code to the environment
#
def addCode (env, filename, code):
  # Add a global "FilesToExtract" attribute to the environment
  if not hasattr(env, 'FilesToExtract'):
    env.FilesToExtract = []
  filenames = [item['filename'] for item in env.FilesToExtract]
  i = 0
  # Prevent multiple files with the same name
  while (filename in filenames):
    i += 1
    filename = '%s.%i' % (filename, i)
  env.FilesToExtract.append({
    'docname': env.docname,
    'filename': filename,
    'content': code,
  })

# Extract the files to extract
#
def extractFiles (app, exception):
  env = app.builder.env
  if hasattr(env, 'FilesToExtract'):
    ext_dir = app.config.extract_dir
    for item in env.FilesToExtract:
      filename = os.path.join(ext_dir, item['filename'])
      directory = os.path.dirname(filename)
      if (not os.path.exists(directory)):
          os.makedirs(directory)
      try:
        with codecs.open(filename, 'w', 'utf-8') as extractedfile:
          extractedfile.write(item['content'])
      except:
        print ('Error creating file %s' % filename)

# Purge the extractfile items from a file when it is changed
#
def purge_FilesToExtract(app, env, docname):
  if hasattr(env, 'FilesToExtract'):
    env.FilesToExtract = [item for item in env.FilesToExtract
                          if item['docname'] != docname]

# Setup
#
def setup(app):
  app.add_directive('extractfile', ExtractFileDirective)
  app.add_directive('extractcode-block', ExtractCodeBlock)
  app.connect('env-purge-doc', purge_FilesToExtract)
  app.connect('build-finished', extractFiles)
  app.add_config_value('extract_dir', 'extracted_files', '')



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later,
  but that will cause
  problems: the code will get executed twice:

  - When you run `python -manalyze_molcas` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``analyze_molcas.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``analyze_molcas.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
from pathlib import Path
import click
from analyze_molcas import SpatialOrbs


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument('input_path', type=click.Path(exists=True))
@click.option('--out', '-o', default=None, type=click.Path(exists=False),
              help='Output file name. '
                   'If omitted, the suffix UhfOrb is appended.')
def spat_to_spin(input_path, out):
    """Transform spatial orbital files to spin orbitals."""
    input_path = Path(input_path)
    if out is None:
        out = input_path.parent / f'{input_path.stem}.UhfOrb'
    SpatialOrbs.read_orbfile(input_path).to_SpinOrbs().write_orbfile(out)


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument('reference_path', type=click.File())
@click.argument('target_path', type=click.File())
def assimilate(reference_path, target_path):
    """Orthogonally transform orbitals in target active space
    as close as possible to reference active space."""
    reference = SpatialOrbs.read_orbfile_stream(reference_path).canonicalize()
    target = SpatialOrbs.read_orbfile_stream(target_path).canonicalize()

    for line in (reference
                 .assimilate(target, [target.get_idx_active_space()])
                 .write_orbfile()):
        print(line)


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument('orb_files', type=click.File(), nargs=-1)
def combine_orbs(orb_files):
    "Combine several RasOrbs files into one. The symmetry group must match."
    import operator
    from functools import reduce

    orbs = (SpatialOrbs.read_orbfile_stream(path) for path in orb_files)

    for line in reduce(operator.add, orbs).write_orbfile():
        print(line)



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015,2017,2019, Ignacio Fdez. Galván                   *
#***********************************************************************

########################################################################
### Patches for pybtex ###

from pybtex.richtext import Text, Protected, String, Tag, BaseMultipartText
from pybtex.scanner import Scanner, Literal, PybtexSyntaxError
from pybtex_docutils import Backend

# Create a new Math element, like Protected but uses a different render
class Math(BaseMultipartText):
  def __init__(self, *args):
    super(Math, self).__init__(*args)
  def capfirst(self):
    return self
  def capitalize(self):
    return self
  def lower(self):
    return self
  def upper(self):
    return self
  def split(self, sep=None, keep_empty_parts=None):
    return [self]
  def __repr__(self):
    reprparts = ', '.join(repr(part) for part in self.parts)
    return 'Math({})'.format(reprparts)
  def render(self, backend):
    text = super(Math, self).render(backend)
    return backend.format_math(text)

# Replace the LaTeX parser to output Math elements too
# (only works with $, and does not handle escaping)
class LaTeXParser(Scanner):
  LBRACE = Literal(u'{')
  RBRACE = Literal(u'}')
  DOLLAR = Literal(u'$')
  def parse(self, level=0):
    return Text(*self.iter_string_parts(level=level))
  def iter_string_parts(self, level=0, in_math=False):
    while True:
      # there will be no Protected inside Math,
      # since we need to preserve all braces
      if in_math:
        token = self.skip_to([self.DOLLAR])
      else:
        token = self.skip_to([self.LBRACE, self.RBRACE, self.DOLLAR])
      if not token:
        remainder = self.get_remainder()
        if remainder:
          yield String(remainder)
        if level != 0:
          raise PybtexSyntaxError('unbalanced braces', self)
        break
      elif token.pattern is self.DOLLAR:
        if in_math:
          yield String(token.value[:-1])
          if level == 0:
            raise PybtexSyntaxError('unbalanced math', self)
          break
        else:
          yield String(token.value[:-1])
          yield Math(*self.iter_string_parts(level=level + 1, in_math=True))
      elif token.pattern is self.LBRACE:
        yield String(token.value[:-1])
        yield Protected(*self.iter_string_parts(level=level + 1))
      else:  # brace.pattern is self.RBRACE
        yield String(token.value[:-1])
        if level == 0:
          raise PybtexSyntaxError('unbalanced braces', self)
        break

@classmethod
def from_latex(cls, latex):
  import codecs
  import latexcodec
  return LaTeXParser(codecs.decode(latex, 'ulatex')).parse()

def format_math(self, text):
  import docutils.nodes
  return [docutils.nodes.math('', '', *text)]

# Patch the class methods we need
Text.from_latex = from_latex
Backend.format_math = format_math

# Function to convert some LaTeX commands
def process_latex(text):
  parts = text.parts
  for i,part in enumerate(parts):
    if isinstance(part, String):
      # Format \emph as Tag('em')
      if (part.endswith(r'\emph ') and isinstance(parts[i+1], Protected)):
        parts[i] = parts[i][:-6]
        parts[i+1] = Tag('em', *parts[i+1].parts)
      # Handle \textsuperscript
      elif (part.endswith(r'\textsuperscript ') and isinstance(parts[i+1], Protected)):
        parts[i] = parts[i][:-17]
        if (str(parts[i+1]) == '2'):
          parts[i+1] = '²'
    elif isinstance(part, Protected):
      parts[i] = process_latex(part)
  if (isinstance(text, Text)):
    return Text(*parts)
  elif (isinstance(text, Protected)):
    return Protected(*parts)

########################################################################

# Set global citation order
citeorder = {}

# Force updating references if anything else changes
def update_bib(app, env, added, changed, removed):
  updated = added.union(changed).union(removed)
  if (app.config.ref_file and (len(updated) > 0)):
    if (app.config.ref_file not in updated):
      return [app.config.ref_file]
  return []

# Put the references file at the end of the list
def bib_at_end(app, env, docnames):
  if (app.config.ref_file):
    try:
      docnames.remove(app.config.ref_file)
      docnames.append(app.config.ref_file)
    except ValueError:
      pass

# Build "citeorder" according to the citation order,
# following the document order in sphinx.
# This has to be done at source-read, because the
# bibliography is sorted before the doctree-read stage
def sort_citations(app, docname, source):
  if (docname == app.config.ref_file):
    env = app.builder.env
    cited = env.bibtex_cache._cited
    rel = env.collect_relations()
    # each element is [up, prev, next],
    # so start with the one with prev=None, and follow next
    doc = [x for x in rel if rel[x][1] is None][0]
    i = 0
    while (doc is not None):
      for cite in cited[doc]:
        if (cite not in citeorder):
          citeorder[cite] = i
          i = i+1
      doc = rel[doc][2]

def setup(app):
  app.connect('env-get-outdated', update_bib)
  app.connect('env-before-read-docs', bib_at_end)
  app.connect('source-read', sort_citations)
  app.add_config_value('ref_file', 'references', 'html')


### Bibliography Styles ###

from pybtex.style.sorting import BaseSortingStyle
from pybtex.style.formatting.unsrt import dashify, date, Style as UnsrtStyle
from pybtex.style.formatting import toplevel
from pybtex.style.template import join, words, field, optional, first_of, names, sentence, tag, optional_field, href
from pybtex.plugin import register_plugin

# Trivial sorting style that just uses "citeorder" as key
class CiteStyle(BaseSortingStyle):
  def sorting_key(self, entry):
    try:
      return citeorder[entry.key]
    except:
      return 9999999

# Formatting style
class MolcasStyle(UnsrtStyle):

  default_sorting_style = 'cite'

  date = words [optional_field('month'), field('year')]

  def format_title(self, e, which_field, as_sentence=True):
    formatted_title = field(
      which_field, apply_func=lambda text: process_latex(text).capitalize()
    )
    if as_sentence:
      return sentence [ formatted_title ]
    else:
      return formatted_title

  def format_btitle(self, e, which_field, as_sentence=True):
    formatted_title = tag('em') [ field(
      which_field, apply_func=lambda text: process_latex(text)
    ) ]
    if as_sentence:
      return sentence [ formatted_title ]
    else:
      return formatted_title

  def format_names(self, role, as_sentence=True):
    formatted_names = names(role, sep=', ', sep2 = ', ', last_sep=', ')
    if as_sentence:
      return sentence [ formatted_names ]
    else:
      return formatted_names

  def get_article_template(self, e):
    pages = first_of [
      # article id with total pages
      optional [
        join [
          field('articleid'),
          optional [
            '(',join[u'1–', optional_field('pagetotal')],')'
          ]
        ]
      ],
      # pages only
      field('pages', apply_func=dashify),
    ]
    volume_and_pages = first_of [
      # volume and pages, with optional issue number
      optional [
        join [
          tag('strong') [field('volume')],
          optional [ '[', field('number'),']' ],
          ' (', field('year'), ')',
          ' ', pages
        ],
      ],
      # pages only
      words ['pages', pages],
    ]
    template = toplevel [
      self.format_names('author'),
      self.format_title(e, 'title'),
      sentence [
        tag('em') [field('journal')],
        optional [ volume_and_pages ],
      ],
      sentence [ optional_field('note') ],
      self.format_web_refs(e),
    ]
    return template

  def get_inbook_template(self, e):
    template = toplevel [
      self.format_author_or_editor(e),
      sentence [
        self.format_btitle(e, 'title', as_sentence=False),
        self.format_chapter_and_pages(e),
      ],
      self.format_volume_and_series(e),
      sentence [
        field('publisher'),
        optional_field('address'),
        optional [
          words [field('edition'), 'edition']
        ],
        date,
        optional_field('note'),
      ],
      self.format_web_refs(e),
    ]
    return template

  def get_incollection_template(self, e):
    template = toplevel [
      sentence [ self.format_names('author') ],
      self.format_title(e, 'title'),
      words [
        'In',
        sentence [
          optional [ self.format_editor(e, as_sentence=False) ],
          self.format_btitle(e, 'booktitle', as_sentence=False),
          self.format_volume_and_series(e, as_sentence=False),
          self.format_chapter_and_pages(e),
        ],
      ],
      sentence [
        optional_field('publisher'),
        optional_field('address'),
        self.format_edition(e),
        date,
      ],
      self.format_web_refs(e),
    ]
    return template

  def get_book_template(self, e):
    template = toplevel [
      self.format_author_or_editor(e),
      self.format_btitle(e, 'title'),
      self.format_volume_and_series(e),
      sentence [
        field('publisher'),
        optional_field('address'),
        self.format_edition(e),
        date
      ],
      optional [ sentence [ self.format_isbn(e) ] ],
      sentence [ optional_field('note') ],
      self.format_web_refs(e),
    ]
    return template

  def get_misc_template(self, e):
    template = toplevel [
      optional [ sentence [ self.format_names('author') ] ],
      optional [ self.format_title(e, 'title') ],
      sentence [
        optional [ field('howpublished') ],
        optional [ date ],
      ],
      sentence [ optional_field('note') ],
      self.format_web_refs(e),
    ]
    return template

  def get_phdthesis_template(self, e):
    template = toplevel [
      sentence [ self.format_names('author') ],
      self.format_btitle(e, 'title'),
      sentence [
        'PhD thesis',
        field('school'),
        optional_field('address'),
        date,
      ],
      sentence [ optional_field('note') ],
      self.format_web_refs(e),
    ]
    return template

  def get_manual_template(self, e):
    template = toplevel [
      optional [ sentence [ self.format_names('author') ] ],
      self.format_btitle(e, 'title'),
      sentence [
        optional_field('organization'),
        optional_field('address'),
        self.format_edition(e),
        optional [ date ],
      ],
      sentence [ optional_field('note') ],
      self.format_web_refs(e),
    ]
    return template

  def format_doi(self, e):
    return href [
      join [
        'https://doi.org/',
        field('doi')
      ],
      join [
        'doi:',
        field('doi')
      ]
    ]

register_plugin('pybtex.style.sorting', 'cite', CiteStyle)
register_plugin('pybtex.style.formatting', 'molcas', MolcasStyle)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Ignacio Fdez. Galván                             *
#***********************************************************************
#
# Parse an output file to get/update the list of used keywords.
# Intended for use with the verify script, to get a list of tested
# keywords, e.g.:
#
# pymolcas verify --postproc "\$MOLCAS_SOURCE/Tools/verify/get_keywords.py \$project.out keylist"

from __future__ import (division, print_function)
import sys, os, re, json

# output file
try:
  outfile = sys.argv[1]
except IndexError:
  sys.exit()

# keywords file
try:
  keyfile = sys.argv[2]
except IndexError:
  sys.exit()

used = {}
try:
  with open(keyfile, 'r') as f:
    used.update(json.load(f))
except IOError:
  pass

val = False
with open(outfile, 'r') as f:
  
  for line in f:
    if line.startswith('-- Input validation'):
      val = True
    elif line.strip() == '--':
      val = False
    if val:
      match = re.match(r'Input for: (\w*)', line)
      if match:
        module = match.group(1)
        if module not in used:
          used[module] = []
      match = re.match(r'\s*> entering group (.*)', line)
      if match:
        group = match.group(1)
        if group not in used[module]:
          used[module].append(group)
      match = re.match(r'\s*keyword (.*)', line)
      if match:
        keyword = match.group(1)
        if keyword not in used[module]:
          used[module].append(keyword)

with open(keyfile, 'w') as f:
  json.dump(used, f, sort_keys=True, indent=2)



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2018, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)
try:
  from builtins import super
except ImportError:
  from future.builtins import super
try:
  from six import text_type, python_2_unicode_compatible
except ImportError:
  text_type = str
  def python_2_unicode_compatible(orig):
    return orig

from os.path import isfile
from re import match
from io import BytesIO

from molcas_aux import *
from tee import teed_call
from simpleeval import simple_eval, SimpleEval

#===============================================================================
# Patch SimpleEval to use the "decimal" module for better precision handling

import ast
import decimal

def _pymolcas_eval(self, expr):
  self.expr = expr
  return text_type(self._eval(ast.parse(expr.strip()).body[0].value))

@staticmethod
def _pymolcas_eval_num(node):
  return decimal.Decimal(text_type(node.n))

SimpleEval.eval = _pymolcas_eval
SimpleEval._eval_num = _pymolcas_eval_num

#===============================================================================

def env_print(env, string):
  if (env.echo):
    print(string)

def check_trap(rc_name=None):
  # this codes should not be affected by MOLCAS_TRAP
  if (rc_name in ['_RC_ALL_IS_WELL_', '_RC_CHECK_ERROR_']):
    return True
  elif (get_utf8('MOLCAS_TRAP', default='on').lower() == 'off'):
    print('*********************************************************')
    print('**                                                     **')
    print('** Non-zero return code, and MOLCAS_TRAP is set to OFF **')
    print('**                                                     **')
    print('*********************************************************')
    return False
  else:
    return True

@python_2_unicode_compatible
class Expression(object):
  def __init__(self, var, op, val):
    self.var = var
    self.op = op
    self.val = val
  def __bool__(self):
    if (self.op == 'eq'):
      return (expandvars(self.var) == expandvars(self.val))
    if (self.op == 'ne'):
      return (expandvars(self.var) != expandvars(self.val))
    if (self.op == 'file'):
      return self.env.file_exists(self.val)
  def __str__(self):
    var = self.var
    if (var == '$MOLCAS_ITER'):
      var = 'ITER'
    if (self.op == 'eq'):
      return '{0} == {1}'.format(var, self.val)
    if (self.op == 'ne'):
      return '{0} != {1}'.format(var, self.val)
    if (self.op == 'file'):
      return '-FILE {0}'.format(self.val)
  #python2
  __nonzero__ = __bool__

class Statement(object):
  def __init__(self, string=None):
    self.rc = None
    self.level = 0
    self.indent = '  '
    if (string is not None):
      self.string = string
  def run(self, env):
    if (env._goto):
      return None
    print('Unsupported statement: {0}'.format(self.string))
    return self.rc

@python_2_unicode_compatible
class Group(Statement):
  def __init__(self, contents, condition=True, rerun=False, var=None, val=None):
    super().__init__()
    self.contents = contents
    self.condition = condition
    self.rerun = rerun
    self.var = var
    self.val = val
    self.maxiter = 1
    self.grouptype = self._find_type()
  def _find_type(self):
    if (isinstance(self.condition, Expression) and (self.rerun is False)):
      return 'if'
    elif ((self.condition is True) and (self.rerun is True)):
      if (self.var is not None):
        return 'foreach'
      else:
        return 'do'
    elif ((self.condition is True) and (self.rerun is False)):
      return 'group'
    else:
      return 'unknown'
  def __str__(self):
    if (self.grouptype == 'group'):
      init = ''
      end = ''
    elif (self.grouptype == 'foreach'):
      init = self.level*self.indent + '>>> FOREACH {0} IN ({1})\n\n'.format(self.var, self.val)
      end = '\n\n' + self.level*self.indent + '>>> END FOREACH'
    elif (self.grouptype == 'do'):
      init = self.level*self.indent + '>>> DO WHILE\n\n'
      end = '\n\n' + self.level*self.indent + '>>> END DO'
    elif (self.grouptype == 'if'):
      if ((len(self.contents) == 1) and isinstance(self.contents[0], Jump)):
        init = self.level*self.indent + '>>> IF ({0}) GOTO {1}'.format(self.condition, self.contents[0].name)
        return init
      init = self.level*self.indent + '>>> IF ({0})\n\n'.format(self.condition)
      end = '\n\n' + self.level*self.indent + '>>> END IF'
    else:
      init = '('
      end = ')'
    return '{0}{1}{2}'.format(init, '\n\n'.join(map(text_type, self.contents)), end)
  def run(self, env):
    if (env._goto):
      return None
    if (self.grouptype == 'foreach'):
      self.values = expandvars(self.val, default='UNKNOWN_VARIABLE')
      re_match = match(r'(-?\d+)\s*\.\.\s*(-?\d+)', self.values)
      if (re_match):
        ini = int(re_match.group(1))
        fin = int(re_match.group(2))
        sign = -1 if (fin < ini) else 1
        self.values = range(ini, fin + sign, sign)
        self.values = list(map(text_type, self.values))
      else:
        self.values = [x.strip() for x in self.values.split(',')]
      self.maxiter = len(self.values)
      env.enter_loop('foreach')
    elif (self.grouptype == 'do'):
      self.previter = get_utf8('MOLCAS_ITER', default='0')
      self.maxiter = int(get_utf8('MOLCAS_MAXITER', default='50'))
      env.enter_loop('do')
    self.thisiter = 1
    if (isinstance(self.condition, Expression)):
      self.condition.env = env
    if (isinstance(self.rerun, Expression)):
      self.rerun.env = env
    while (self.thisiter <= self.maxiter):
      if (self.grouptype == 'if'):
        env_print(env, '\n>>> IF ({0})'.format(self.condition))
      itercontents = self.contents[:]
      if (self.condition):
        i = 0
        if (self.grouptype == 'foreach'):
          Assignment(self.var, self.values[self.thisiter-1]).run(env)
        elif (self.grouptype == 'do'):
          set_utf8('MOLCAS_ITER', self.thisiter)
        no_break = True
        if (self.grouptype == 'do'):
          rc_name = '_RC_CONTINUE_LOOP_'
        else:
          rc_name = '_RC_ALL_IS_WELL_'
        while ((i < len(itercontents)) and (no_break)):
          item = itercontents[i]
          item.group = itercontents
          rc = item.run(env)
          # Transparent return: only update return code if not None
          if (rc is not None):
            self.rc = rc
            rc_name = env.rc_to_name(self.rc)
          no_break = rc_name in ['_RC_ALL_IS_WELL_', '_RC_CONTINUE_LOOP_', '_RC_CONTINUE_UNIX_LOOP_']
          # check MOLCAS_TRAP here only if it's not the last element in the group
          if (i+1 < len(itercontents)):
            if ((not no_break) and (not check_trap(rc_name))):
              no_break = True
              rc_name = '_RC_ALL_IS_WELL_'
          set_utf8('EMIL_RETURNCODE', 0 if self.rc is None else self.rc)
          i += 1
      else:
        env_print(env, '(Skipped)')
      # Termination conditions
      # TODO: should MOLCAS_TRAP be checked here too?
      # TODO: run EMIL commands after a module fails (see FS#37)
      if (env._goto):
        break
      # this seems to be needed (here) for compatibility with old driver
      if (self.grouptype in ['do', 'foreach']):
        env._check_count += 1
      if (not self.rerun):
        break
      # in a DO loop, "all is well" means the loop terminates
      if (self.grouptype == 'do'):
        no_break = rc_name in ['_RC_CONTINUE_LOOP_', '_RC_CONTINUE_UNIX_LOOP_']
      elif (self.grouptype == 'foreach'):
        no_break = rc_name in ['_RC_ALL_IS_WELL_', '_RC_CONTINUE_LOOP_', '_RC_CONTINUE_UNIX_LOOP_']
      else:
        no_break = rc_name in ['_RC_ALL_IS_WELL_']
      # no-trap and loops are tricky
      if ((not no_break) and (not check_trap(rc_name))):
        if ((rc_name not in ['_RC_NOT_CONVERGED_']) or (self.grouptype != 'do')):
          no_break = True
          rc_name = '_RC_ALL_IS_WELL_'
      if (self.grouptype == 'do'):
        env_print(env, '\n>>> END DO')
      elif (self.grouptype == 'foreach'):
        env_print(env, '\n>>> END FOREACH')
      if (not no_break):
        break
      self.thisiter += 1
    if (self.grouptype == 'do'):
      if (not env._goto):
        if (rc_name in ['_RC_CONTINUE_LOOP_', '_RC_CONTINUE_UNIX_LOOP_']):
          self.rc = '_RC_NOT_CONVERGED_'
      set_utf8('MOLCAS_ITER', self.previter)
      env.exit_loop()
    if (self.grouptype == 'foreach'):
      env.exit_loop()
    set_utf8('EMIL_RETURNCODE', self.rc)
    return self.rc

@python_2_unicode_compatible
class Assignment(Statement):
  def __init__(self, var, val, literal=True):
    super().__init__()
    self.literal = literal
    self.var = var
    self.val = val
  def __str__(self):
    if (self.literal):
      return self.level*self.indent + '>>> EXPORT {0} = {1}'.format(self.var, self.val)
    else:
      return self.level*self.indent + '>>> EVAL {0} = {1}'.format(self.var, self.val)
  def run(self, env):
    if (env._goto):
      return None
    if (self.literal):
      val = expandvars(self.val, default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> EXPORT {0} = {1}'.format(self.var, val))
      set_utf8(self.var, val)
    else:
      val = expandvars(self.val, default='0')
      try:
        eval_val = simple_eval(val)
      except:
        eval_val = ''
      env_print(env, '\n>>> EVAL {0} = {1} = {2}'.format(self.var, val, eval_val))
      set_utf8(self.var, eval_val)
    return self.rc

@python_2_unicode_compatible
class Break(Statement):
  def __init__(self, rc):
    super().__init__()
    self.rc = rc
  def __str__(self):
    num = ' {0}'.format(self.rc)
    if (self.rc is None):
      num = ''
    return self.level*self.indent + '>>> EXIT{0}'.format(num)
  def run(self, env):
    if (env._goto):
      return None
    try:
      num = expandvars(self.rc)
      rc = int(num)
    except:
      rc = None
    if (rc is None):
      rc = '_RC_EXIT_'
    if (rc == '0'):
      rc = '_RC_EXIT_EXPECTED_'
    env_print(env, '\n>>> EXIT {0}'.format(env.rc_to_name(rc)))
    return rc

@python_2_unicode_compatible
class Include(Statement):
  def __init__(self, filename):
    super().__init__()
    self.filename = filename
  def __str__(self):
    return self.level*self.indent + '>>> INCLUDE {0}'.format(self.filename)
  def run(self, env):
    if (env._goto):
      return None
    # NOTE: circular dependency here
    import emil_parse
    pos = self.group.index(self)+1
    fname = expandvars(self.filename, default='UNKNOWN_VARIABLE')
    env_print(env, '\n>>> INCLUDE {0}'.format(fname))
    if (isfile(fname)):
      blocks, files = emil_parse.EMIL_Parse(fname)
      env.add_files(files)
      self.group.insert(pos, blocks)
    else:
      self.rc = '_RC_INPUT_EMIL_ERROR_'
    return self.rc

@python_2_unicode_compatible
class Program(Statement):
  def __init__(self, lines):
    super().__init__()
    self.lines = lines
  def __str__(self):
    name = '&' + self.lines[1].upper()
    code = ('\n' + self.level*self.indent).join(self.lines[2:])
    if (code != ''):
      code = '\n' + self.level*self.indent + code
    return self.level*self.indent + name + code
  def run(self, env):
    if (env._goto):
      return None
    name = self.lines[1]
    inp = expandvars('\n'.join(self.lines[2:]), default='UNKNOWN_VARIABLE')
    rc = env.run_module(name, inp)
    if (rc is not None):
      self.rc = rc
    return self.rc

# TODO: convert to ParTask?
@python_2_unicode_compatible
class System(Statement):
  def __init__(self, commandline, parallel=False):
    super().__init__()
    self.commandline = commandline
    self.parallel = parallel
  @property
  def expanded_commandline(self):
    return expandvars(self.commandline, skip_escaped=True)
  @property
  def name(self):
    if (self.parallel):
      return "EXEC"
    else:
      return "SHELL"
  def __str__(self):
    return self.level*self.indent + '>>> {0} {1}'.format(self.name, self.commandline)
  def run(self, env):
    if (env._goto):
      return None
    output = BytesIO()
    error = BytesIO()
    env_print(env, '\n>>> {0} {1}'.format(self.name, self.expanded_commandline))
    if (env.allow_shell):
      if (self.parallel):
        task = ['!'] + self.expanded_commandline.split()
        self.rc = env.parallel_task(task)
      else:
        self.rc = teed_call(self.expanded_commandline, shell=True, cwd=env.scratch, stdout=output, stderr=error)
    else:
      env_print(env, '(Shell commands disabled)')
    # Successful return is transparent: it inherits the previous return code
    if (self.rc == 0):
      self.rc = None
    return self.rc

@python_2_unicode_compatible
class Setting(Statement):
  def __init__(self, var, val):
    super().__init__()
    self.var = var
    self.val = val
  def __str__(self):
    if (self.var == 'echo'):
      val = self.val
      if (val in ['on', 'off']):
        val = val.upper()
      return self.level*self.indent + '>>> ECHO {0}'.format(val)
  def run(self, env):
    if (env._goto):
      return None
    if (self.var == 'echo'):
      if (self.val == 'on'):
        env.echo = True
      elif (self.val == 'off'):
        env.echo = False
      else:
        print(self.val)
    return self.rc

@python_2_unicode_compatible
class Label(Statement):
  def __init__(self, name):
    super().__init__()
    self.name = name
  def __str__(self):
    return self.level*self.indent + '>>> LABEL {0}'.format(self.name)
  def run(self, env):
    if (env._goto):
      if (env._goto == self.name):
        env._goto = False
    return self.rc

@python_2_unicode_compatible
class Jump(Statement):
  def __init__(self, name):
    super().__init__()
    self.name = name
  def __str__(self):
    return self.level*self.indent + '>>> GOTO {0}'.format(self.name)
  def run(self, env):
    if (env._goto):
      return None
    env_print(env, '\n>>> GOTO {0}'.format(self.name))
    env._goto = self.name
    return self.rc

@python_2_unicode_compatible
class ParTask(Statement):
  def __init__(self, action, force, args):
    super().__init__()
    self.action = action
    self.force = force
    self.args = args
  def __str__(self):
    f = '-FORCE ' if self.force else ''
    if (self.action == 'rm'):
      return self.level*self.indent + '>>> RM {0}{1}'.format(f, self.args[0])
    if (self.action == 'copy'):
      return self.level*self.indent + '>>> COPY {0}{1} {2}'.format(f, self.args[0], self.args[1])
    if (self.action == 'save'):
      return self.level*self.indent + '>>> SAVE {0}{1} {2}'.format(f, self.args[0], self.args[1])
    if (self.action == 'clone'):
      return self.level*self.indent + '>>> CLONE {0}{1} {2}'.format(f, self.args[0], self.args[1])
    if (self.action == 'collect'):
      return self.level*self.indent + '>>> COLLECT {0}{1} {2}'.format(f, self.args[0], self.args[1])
  def run(self, env):
    if (env._goto):
      return None
    f = '-FORCE ' if self.force else ''
    if (self.action == 'rm'):
      fname = expandvars(self.args[0], default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> RM {0}{1}'.format(f, fname))
      task = ['x', fname]
    if (self.action == 'copy'):
      f_in = expandvars(self.args[0], default='UNKNOWN_VARIABLE')
      f_out = expandvars(self.args[1], default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> COPY {0}{1} {2}'.format(f, f_in, f_out))
      task = ['c', '1', f_in, f_out]
    if (self.action == 'save'):
      f_in = expandvars(self.args[0], default='UNKNOWN_VARIABLE')
      f_out = expandvars(self.args[1], default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> SAVE {0}{1} {2}'.format(f, f_in, f_out))
      task = ['c', '0', f_in, f_out]
    if (self.action == 'clone'):
      f_in = expandvars(self.args[0], default='UNKNOWN_VARIABLE')
      f_out = expandvars(self.args[1], default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> CLONE {0}{1} {2}'.format(f, f_in, f_out))
      task = ['c', '3', f_in, f_out]
    if (self.action == 'collect'):
      f_in = expandvars(self.args[0], default='UNKNOWN_VARIABLE')
      f_out = expandvars(self.args[1], default='UNKNOWN_VARIABLE')
      env_print(env, '\n>>> COLLECT {0}{1} {2}'.format(f, f_in, f_out))
      task = ['c', '2', f_in, f_out]
    self.rc = env.parallel_task(task, force=self.force)
    if (self.force):
      self.rc = 0
    # Successful return is transparent: it inherits the previous return code
    if (self.rc == 0):
      self.rc = None
    else:
      self.rc = '_RC_INPUT_EMIL_ERROR_'
    return self.rc

@python_2_unicode_compatible
class Python(Statement):
  '''
  This is not really part of the "abstract flow" family, but it's included
  for consistency. Intended to be used with arbitrary python code, instead
  of some specific input language (like EMIL).
  '''
  def __init__(self, filename):
    self.rc = None
    with utf8_open(filename) as f_in:
      self.text = f_in.read()
      # If there's an error in the code SyntaxError will be raised
      # and catched upstream
      self.code = compile(self.text, filename, 'exec')
  def __str__(self):
    return self.text.rstrip()
  def run(self, env):
    exec(self.code, {'MOLCAS': env})
    self.rc = 0
    return self.rc



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2018, Ignacio Fdez. Galván                             *
#***********************************************************************

import os
import errno

def read_answer(answers, retry):
  while True:
    try:
      ans = raw_input()
    except NameError:
      ans = input()
    ans = ans.lower()
    if (ans == ''):
      return ans
    elif (ans in answers):
      return answers[ans]
    else:
      print(retry)

def read_number():
  while True:
    try:
      ans = raw_input()
    except NameError:
      ans = input()
    if (ans == ''):
      return ans
    try:
      ans = int(ans)
    except:
      ans = -1
    if (ans > 0):
      return ans
    else:
      print('Please enter a positive number')

def read_string():
  while True:
    try:
      ans = raw_input()
    except NameError:
      ans = input()
    if (' ' not in ans):
      return ans
    else:
      print('It is a bad idea to use spaces in the path')

def write_molcasrc(molcasrc, program='molcas'):
  config = []

  print("\n"
  "This function will help you create a custom configuration file (molcasrc) for\n"
  "OpenMolcas. The file contains settings for some useful environment variables,\n"
  "you can always modify the file, or override any environment variable for\n"
  "specific calculations.")

  print("\n"
  "MOLCAS_PROJECT\n"
  "--------------\n"
  "Specifies how the project name is set if the \"Project\" variable is not defined.\n"
  "\n"
  "Enter a number, or press enter to keep the default (1):\n"
  "1) NAME:    Use the input filename (minus extension) as project name.\n"
  "2) NAMEPID: Same as 1, but also use a unique name for the scratch directory.\n"
  "3) TIME:    Use the current time as a project name.")
  ans = read_answer({'1': 'NAME', '2': 'NAMEPID', '3': 'TIME'}, 'Please enter 1, 2 or 3')
  if (ans != ''):
    config.append('MOLCAS_PROJECT={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_MEM\n"
  "----------\n"
  "Defines the approximate maximum memory used by OpenMolcas during a calculation.\n"
  "\n"
  "Enter a number (in MiB), or press enter to keep the installation default:")
  ans = read_number()
  if (ans != ''):
    config.append('MOLCAS_MEM={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_MAXITER\n"
  "--------------\n"
  "Defines the maximum number of iterations for a \"Do While\" loop in the input.\n"
  "\n"
  "Enter a number, or press enter to keep the default (50):")
  ans = read_number()
  if (ans != ''):
    config.append('MOLCAS_MAXITER={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_NEW_DEFAULTS\n"
  "-------------------\n"
  "Specify whether or not to use the new OpenMolcas defaults, in particular,\n"
  "with new defaults enabled:\n"
  "\n"
  "* RICD will be enabled (disable with NOCD)\n"
  "* IPEA shift will be disabled (set previous default with IPEA=0.25)\n"
  "\n"
  "Enter y or n, or press enter to keep the default (n):\n"
  "y) YES: The new defaults will be used.\n"
  "n) NO:  The previous defaults will be kept.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_NEW_DEFAULTS={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_TRAP\n"
  "-----------\n"
  "Specifies whether a calculation should stop when a module reports a failure.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) YES: The calculation stops after a failure.\n"
  "n) NO:  The calculation continues after a failure.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_TRAP={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_WORKDIR\n"
  "--------------\n"
  "Defines the parent directory inside which a scratch directory for each\n"
  "calculation will be created. It can be an absolute path, a relative (with\n"
  "respect to the submit directory) path, or the special value \"PWD\", which is\n"
  "equivalent to using \".\" as relative path.\n"
  "\n"
  "Enter a string, or press enter to keep the system's default ($TMPDIR):")
  ans = read_string()
  if (ans != ''):
    config.append('MOLCAS_WORKDIR={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_NEW_WORKDIR\n"
  "------------------\n"
  "Specifies whether the scratch directory will be cleaned *before* a\n"
  "calculation. This can also be specified with the -new and -old flags\n"
  "for {}.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (n):\n"
  "y) YES: The scratch directory is cleaned before a calculation.\n"
  "n) NO:  The scratch directory is not cleaned before a calculation.".format(program))
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_NEW_WORKDIR={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_KEEP_WORKDIR\n"
  "-------------------\n"
  "Specifies whether the scratch directory will be cleaned *after* a calculation.\n"
  "This can also be overridden with the -clean flag for {}.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) YES: The scratch directory is kept after a calculation.\n"
  "n) NO:  The scratch directory is cleaned after a calculation.".format(program))
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_KEEP_WORKDIR={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_OUTPUT\n"
  "-------------\n"
  "Specifies where to save generated files, like orbitals and molden files.\n"
  "It can be an absolute or relative (with respect to the submit directory) path,\n"
  "or the special values \"NAME\" (a subdirectory will be created with the name of\n"
  "the project) or \"PWD\" (files will be saved in the submit directory).\n"
  "\n"
  "Enter a string, or press enter to keep the default (PWD):")
  ans = read_string()
  if (ans != ''):
    config.append('MOLCAS_OUTPUT={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_SAVE\n"
  "-----------\n"
  "Defines how to alter filenames to avoid overwriting existing files.\n"
  "\n"
  "Enter a number or, press enter to keep the dafault (1):\n"
  "1) REPL: Overwrite existing files.\n"
  "2) ORIG: Rename existing files to add the extension \".orig\".\n"
  "3) INCR: New files will have a extension with incremental numbers.")
  ans = read_answer({'1': 'REPL', '2': 'ORIG', '3': 'INCR'}, 'Please enter 1, 2 or 3')
  if (ans != ''):
    config.append('MOLCAS_SAVE={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_MOLDEN\n"
  "-------------\n"
  "Specifies whether Molden format files with molecular orbitals will be created.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) ON:  Molden orbital files are created.\n"
  "n) OFF: Molden orbital files are not created.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_MOLDEN={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_PRINT\n"
  "------------\n"
  "Defines the default print level for OpenMolcas.\n"
  "\n"
  "Enter a number, or press enter to keep the default (2):\n"
  "0) SILENT\n"
  "1) TERSE\n"
  "2) NORMAL\n"
  "3) VERBOSE\n"
  "4) DEBUG\n"
  "5) INSANE")
  ans = read_answer({'0': 'SILENT', '1': 'TERSE', '2': 'NORMAL', '3': 'VERBOSE', '4': 'DEBUG', '5': 'INSANE'}, 'Please enter 0, 1, 2, 3, 4 or 5')
  if (ans != ''):
    config.append('MOLCAS_PRINT={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_ECHO_INPUT\n"
  "-----------------\n"
  "Determines if the pre-processed input will be included in the output.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) YES: Include the input at the beginning of the output.\n"
  "n) NO:  Do not include the input in the output.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_ECHO_INPUT={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_COLOR\n"
  "------------\n"
  "Specifies whether to use a simple markup for important information in the\n"
  "output. This markup can be displayed by e.g. vim (see the Tools/syntax\n"
  "directory).\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) YES: Simple markup will be added.\n"
  "n) NO:  No markup will be added.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_COLOR={}'.format(ans))
    print('')

  print("\n"
  "MOLCAS_REDUCE_PRT\n"
  "-----------------\n"
  "Specifies whether the print level will be reduced inside a \"Do While\" loop.\n"
  "\n"
  "Enter y or n, or press enter to keep the default (y):\n"
  "y) YES: Print level is reduced inside a \"Do While\" loop.\n"
  "n) NO:  Print level is the same as outside.")
  ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
  if (ans != ''):
    config.append('MOLCAS_REDUCE_PRT={}'.format(ans))
    print('')

  if (config):
    print("\n"
    "Based on your answers, we recommend the following molcasrc file:\n"
    "-------------------------")
    print("\n".join(config))
    print('-------------------------')

    print("\n"
    "The file {} will be created or overwritten.\n"
    "Is this OK? (y/N)".format(molcasrc))
    ans = read_answer({'y': 'YES', 'n': 'NO'}, 'Please answer y or n')
    if (ans == 'YES'):
      try:
        dirrc = os.path.dirname(molcasrc)
        if (not os.path.exists(dirrc)):
          try:
            os.makedirs(dirrc)
          except OSError as e:
            if (e.errno != errno.EEXIST):
              raise
        with open(molcasrc, 'w') as f:
          f.write("# molcasrc configuration file written by {} -setup\n\n".format(program))
          f.write("\n".join(config))
          f.write("\n")
      except:
        raise
        return False
  else:
    print("\n"
    "All defaults were selected, no molcasrc file will be written.")

  return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#***********************************************************************

from __future__ import print_function
import fileinput
import re

doxygen_re = re.compile(r"^[*Cc!]>")
link_re = re.compile(r"::(\w*)")

# Replace [::FuncName] with [\ref funcname "FuncName"]
reflink = lambda pat: r'\ref {} "{}"'.format(pat.group(1).lower(), pat.group(1))

for line in fileinput.input():
  if (doxygen_re.search(line)):
    line = re.sub(link_re,reflink,line)
    print(line,end='')
  else:
    print(line,end='')



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

from typing import Sequence, TextIO
from os import PathLike
from copy import deepcopy

import numpy as np
from numpy import argsort, isclose, reshape, argwhere
from scipy.linalg import block_diag

from analyze_molcas._base_orbfile import (
    _Orbitals, FileFormat, _reindex,
    _read_spatial_orbs, _read_header,
    _write_section_1D, _write_section_2D)

from analyze_molcas.spin_orbfile import SpinOrbs

from analyze_molcas._linalg_helper import get_kabsch_transformation


class SpatialOrbs(_Orbitals):
    """Represents a RasOrb file with spatial orbitals

The operators +, *, ==, and != are overloaded.
Similar to lists `+` concatenates two SpatialOrbs and
`*` concatenates a SpatialOrb with itself n-times.


Attributes
----------
All attributes are lists, whose length is the number of
irreps.  The i-th element is the i-th irrep.

orbs : list of integer
    The number of orbitals per irrep.
    `RasOrb.orb[i]` is the number of orbitals in the i-th irrep.
    `sum(RasOrb.orb)` is the total number of orbitals.
coeff : list of float 2D arrays
    The MO-coefficients per irrep.
    `RasOrb.coeff[i][mu, l]` is the linear coefficient of the
    mu-th AO in the l-th MO in the i-th irrep.
occ : list of float
    The occupation numbers per irrep.
    `RasOrb.occ[i][j]` is the occupation number of the j-th MO
    in the i-th irrep.
energy : list of float
    The orbital energies per irrep.
    `RasOrb.energy[i][j]` is the energy of the j-th MO
    in the i-th irrep.
idx : list of string 1D arrays
    The RAS index per irrep.
    `RasOrb.energy[i][j]` is the RAS index of the j-th MO
    in the i-th irrep.

Examples
--------
Read a RasOrb file:

>>> ras_orb = SpatialOrbs.read_orbfile('./quint.RasOrb')

Write a RasOrb file:

>>> ras_orb.write_orbfile('./transformed.RasOrb')

Put all orbitals in active space:

>>> for index in ras_orb.idx:
>>>     index[:] = '2'
    """

    @classmethod
    def read_orbfile(cls, path: PathLike):
        with open(path, 'r') as f:
            return cls.read_orbfile_stream(f)

    @classmethod
    def read_orbfile_stream(cls, stream: TextIO):
        version, orbs, spin_orbs = _read_header(stream)

        if spin_orbs:
            raise FileFormat('File represents spin orbitals')

        coeff, occ, energy, idx = _read_spatial_orbs(stream, orbs, version)
        return cls(orbs, coeff, occ, energy, idx)

    def reindex(self,
                new_idx: Sequence[Sequence[int]], inplace: bool=False):
        if inplace:
            self.coeff = [
                coeff[:, idx] for idx, coeff in zip(new_idx, self.coeff)]
            self.energy = _reindex(self.energy, new_idx)
            self.idx = _reindex(self.idx, new_idx)
            self.occ = _reindex(self.occ, new_idx)
        else:
            new = self.copy()
            new.reindex(new_idx, inplace=True)
            return new

    def round_occ(self, inplace: bool=False):
        new_occ = [occ.round() for occ in self.occ]

        if not isclose(sum(occ.sum() for occ in self.occ),
                       sum(occ.sum() for occ in new_occ)):
            raise ValueError('Rounded occupation numbers yield different '
                             'number of electrons.')
        if inplace:
            self.occ = new_occ
        else:
            new = self.copy()
            new.round_occ(inplace=True)
            return new

    def to_SpinOrbs(self) -> SpinOrbs:
        spat = self.reindex([argsort(-v, kind='stable') for v in self.occ])

        new_occ = {'a': [occ / 2.0 for occ in spat.occ],
                   'b': [occ / 2.0 for occ in spat.occ]}

        return SpinOrbs(
            orbs=spat.orbs.copy(), coeff=self._spin_copy(spat.coeff),
            occ=new_occ, energy=self._spin_copy(spat.energy),
            idx=spat.idx.copy())

    @staticmethod
    def _spin_copy(v):
        return {'a': deepcopy(v), 'b': deepcopy(v)}

    def _write_header(self):
        return ['#INPORB 2.2',
                '#INFO',
                '* PyOrb written',
                f"{0:8d}{len(self.orbs):8d}{0:8d}",
                f"{''.join(f'{x:8d}' for x in self.orbs)}",
                f"{''.join(f'{x:8d}' for x in self.orbs)}",
                '* SpatialOrbs written from analyze_molcas']

    def _write_MO_coeff(self):
        def get_paragraph(irrep, orb):
            return f'* ORBITAL{irrep + 1:5d}{orb + 1:5d}'

        yield from _write_section_2D(
            title='#ORB', paragraph=get_paragraph,
            orbs=self.orbs, values=self.coeff)

    def _write_MO_occ(self):
        yield from _write_section_1D(
            title='#OCC', subtitle='* OCCUPATION NUMBERS',
            orbs=self.orbs, values=self.occ)

    def _write_MO_human_occ(self):
        yield from _write_section_1D(
            title='#OCHR', subtitle='* OCCUPATION NUMBERS (HUMAN-READABLE)',
            orbs=self.orbs, values=self.occ, cols=10, fmt='8.4f')

    def _write_MO_E(self):
        yield from _write_section_1D(
            title='#ONE', subtitle='* ONE ELECTRON ENERGIES',
            orbs=self.orbs, values=self.energy, cols=10, fmt='12.4E')

    def canonicalize(self):
        """Sort into canonical MOLCAS order."""
        def to_key(v):
            ORB_INDEX = ['f', 'i', '1', '2', '3', 's', 'd']
            D = {index: i for i, index in enumerate(ORB_INDEX)}
            return np.array([D[i] for i in v], dtype='i8')

        idx = [argsort(to_key(idx), kind='stable') for idx in self.idx]
        new = self.reindex(idx)
        return new

    def get_idx_active_space(self):
        return [reshape(argwhere(condition), sum(condition))
                for condition in (index == '2' for index in self.idx)]

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("Can only add {}".format(self.__class__))
        if len(self.orbs) != len(other.orbs):
            raise ValueError("Symmetry group does not match.")
        return self.__class__(
            orbs=[n_s + n_o for n_s, n_o in zip(self.orbs, other.orbs)],
            coeff=[block_diag(Q, P) for Q, P in zip(self.coeff, other.coeff)],
            occ=[cat(q, p) for q, p in zip(self.occ, other.occ)],
            energy=[cat(q, p) for q, p in zip(self.energy, other.energy)],
            idx=[cat(q, p) for q, p in zip(self.idx, other.idx)])

    def __radd__(self, other):
        return self + other

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("Can't multiply Orbfile by non-int.")
        if other < 0:
            raise ValueError("Other has to be >=0")

        def block_times(n, A):
            return block_diag(*(A for _ in range(n)))

        def cat_times(n, v):
            return np.concatenate([v for _ in range(n)], axis=0)

        n = other
        return self.__class__(
            orbs=[n * n_orbs for n_orbs in self.orbs],
            coeff=[block_times(n, Q) for Q in self.coeff],
            occ=[cat_times(n, q) for q in self.occ],
            energy=[cat_times(n, q) for q in self.energy],
            idx=[cat_times(n, q) for q in self.idx])

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        if not self._same_dims(other):
            return False
        return all([
            (Q == P for Q, P in zip(self.coeff, other.coeff)),
            (q == p for q, p in zip(self.occ, other.occ)),
            (q == p for q, p in zip(self.energy, other.energy)),
            (q == p for q, p in zip(self.idx, other.idx))])

    def __neq__(self, other):
        return not self == other

    def _same_dims(self, other):
        if len(self.orbs) != len(other.orbs):
            return False
        if any(n_s != n_o for n_s, n_o in zip(self.orbs, other.orbs)):
            return False
        return True


    def assimilate(self, other, blocks):
        """Make other blockwise similar to self

    Parameters
    ----------

    other : SpatialOrb
    blocks : list of list of indices
        `blocks[i][j]` contains the index of the j-th block in the i-th irrep
        """
        if not self._same_dims(other):
            raise ValueError("Self and Other have to be same symmetry group "
                             "with the same number of orbitals per irrep")
        new_other = other.copy()
        # Defines a shorter function name
        f = get_kabsch_transformation
        for irrep, indices in enumerate(blocks):
            for GAS_space in indices:
                R = f(self.coeff[irrep][:, GAS_space],
                      other.coeff[irrep][:, GAS_space])
                new_other.coeff[irrep][:, GAS_space] = (
                    other.coeff[irrep][:, GAS_space] @ R.T)
        return new_other


def cat(*args):
    return np.concatenate(args, axis=0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#***********************************************************************
# Lots of problems with this script!!!

from __future__ import print_function
from __future__ import unicode_literals
import os.path
import sys
import re
import xml.etree.ElementTree as ET
import htmlentitydefs

basedir = 'xml'
extension = '.xml'
htmldir = 'html'
htmlextension = '.xhtml'

# Read argument
try:
  argument = sys.argv[1]
except IndexError:
  argument = None

def in_red(text):
  return '\033[31m{}\033[0m'.format(text)

def in_brown(text):
  return '\033[33m{}\033[0m'.format(text)

def in_blue(text):
  return '\033[34m{}\033[0m'.format(text)

database = ET.parse(os.path.join(basedir,'index'+extension))
root = database.getroot()

# First create a list of files, containing the directory where each find is located.
file_list={}
# Iterate directory items
for diritem in root.iterfind('compound[@kind="dir"]'):
  dirname = diritem.findtext('name')
  # Open the XML file for each directory
  dirxml = diritem.get('refid') + extension
  dirtree = ET.parse(os.path.join(basedir,dirxml))
  # For each file in this directory, assign the director name
  for fileitem in dirtree.getroot().iter('innerfile'):
    file_list[fileitem.get('refid')] = dirname

# Create a dictionary linking each element with its parent
parent_map = {c:p for p in root.iter() for c in p}

# Build the dictionary of functions, containg the file where each function is defined
name_list={}
# Iterate all "function" members belonging to "file" compounds
for member in root.iterfind('compound[@kind="file"]/member[@kind="function"]'):
  name = member.findtext('name')
  refid = member.get('refid')
  # Find the directory and filename
  parent = parent_map[member]
  filename = parent.findtext('name')
  fileref = parent.get('refid')
  # Ignore .h files
  if (not re.search(r"\.h$",filename)):
    dirname = file_list[parent.get('refid')]
    filename = os.path.join(dirname,filename)
    # Add an entry to the dictionary
    name_list.setdefault(name,[]).append([filename,fileref,refid])

#===============================================================================

def get_info(fileref,refid):
  import HTMLParser
  h = HTMLParser.HTMLParser()
  filedata = ET.parse(os.path.join(basedir,fileref+extension))
  root = filedata.getroot()
  memberdef = root.find('.//memberdef[@id="{}"]'.format(refid))
  # Insert Unicode characters
  for item in memberdef.findall('.//grave'):
    item.text = h.unescape('&{}grave;'.format(item.get('char')))
  for item in memberdef.findall('.//acute'):
    item.text = h.unescape('&{}acute;'.format(item.get('char')))
  for item in memberdef.findall('.//circ'):
    item.text = h.unescape('&{}circ;'.format(item.get('char')))
  for item in memberdef.findall('.//tilde'):
    item.text = h.unescape('&{}tilde;'.format(item.get('char')))
  for item in memberdef.findall('.//umlaut'):
    item.text = h.unescape('&{}uml;'.format(item.get('char')))
  for item in memberdef.findall('.//ring'):
    item.text = h.unescape('&{}ring;'.format(item.get('char')))
  for item in memberdef.findall('.//cedil'):
    item.text = h.unescape('&{}cedil;'.format(item.get('char')))
  for item in memberdef.findall('.//slash'):
    item.text = h.unescape('&{}slash;'.format(item.get('char')))
  for item in memberdef.findall('.//larr'):
    item.text = h.unescape('&larr;')
  for item in memberdef.findall('.//rarr'):
    item.text = h.unescape('&rarr;')
  for item in memberdef.findall('.//times'):
    item.text = h.unescape('&times;')
  for item in memberdef.findall('.//le'):
    item.text = h.unescape('&le;')
  for item in memberdef.findall('.//ge'):
    item.text = h.unescape('&ge;')
  for item in memberdef.findall('.//ne'):
    item.text = h.unescape('&ne;')
  for item in memberdef.findall('.//sp'):
    item.text = ' '
  # Find detailed section
  detailednode = memberdef.find('detaileddescription')
  if (len(''.join(detailednode.itertext()).strip()) == 0): detailednode = memberdef.find('inbodydescription')
  # Process definition
  definition = memberdef.findtext('definition')
  argsstring = memberdef.findtext('argsstring')
  print(' {}: {}{}'.format(in_blue('Definition'),definition,argsstring))
  # Process arguments
  parameterlist = memberdef.find('.//parameterlist')
  params = []
  size = 0
  # Get parameter definitions from XML
  if (parameterlist is not None):
    for item in parameterlist.iterfind('parameteritem'):
      names = []
      for param in item.iterfind('parameternamelist/parametername'):
        if (param.get('direction') is None):
          names.append(param.text)
        else:
          names.append('{} [{}]'.format(param.text,param.get('direction')))
      names = ', '.join(names)
      description = ''.join(item.find('parameterdescription').itertext())
      description = description.strip()
      description = re.sub(r"\n",' ',description)
      params.append([names,description])
      size = max(size,len(names))
  # Get only parameter names from XML
  else:
    for item in memberdef.iterfind('param'):
      name = item.findtext('defname')
      if (name is None): name = item.findtext('declname')
      params.append([name,''])
      size = max(size,len(name))
  # Get parameter types from HTML (due to bug in doxygen's XML output)
  newrefid = re.match(r"(.*)_",refid).group(1)
  parser = ET.XMLParser()
  parser.parser.UseForeignDTD(True)
  parser.entity.update(htmlentitydefs.entitydefs)
  filedata = ET.parse(os.path.join(htmldir,newrefid+htmlextension), parser)
  root = filedata.getroot()
  # Since some parameters may be listed together, we need some kind of index
  index = []
  for i in range(len(params)):
    for j in range(params[i][0].count(', ')+1):
      index.append(i)
  # The types are read from the XML and added to the description
  i = 0
  for item in root.findall('.//*[@class="memproto"]//*[@class="paramtype"]'):
    typ = re.sub(r"(, $|\xa0)",'',''.join(item.itertext()))
    # Define implicit types (assuming default implicit)
    if (len(typ) == 0):
      if (re.match(r"[ijklmn]",params[i][0][0], flags=re.IGNORECASE)):
        typ = '(integer)?'
      else:
        typ = '(real*8)?'
    params[index[i]][1] += '  {}'.format(in_brown(typ))
    i += 1
  # Print arguments
  print(' {}:'.format(in_blue('Arguments')))
  formatstring = '  {{:{}s}}: {{}}'.format(size)
  for (names,description) in params:
    print(formatstring.format(names,description))
  print()
  # Process brief description
  briefdescription = ''.join(memberdef.find('briefdescription').itertext()).strip()
  if (len(briefdescription) > 0):
    print(' {}:'.format(in_blue('Brief description')))
    print('  {}'.format(briefdescription))
  # Process authors
  authors = []
  for item in memberdef.findall('.//simplesect[@kind="author"]/para'):
    authors.append(''.join(item.itertext()).strip())
  for i in range(len(authors)):
    if (i == 0):
      print(' {}: {}'.format(in_blue('Author'),authors[i]))
    else:
      print('         {}'.format(authors[i]))
  if (len(authors) > 0): print()
  # Remove unwanted items from the detailed description
  for item in detailednode.iter():
    if (item.tag in ('parameterlist','simplesect')):
      tail = item.tail
      item.clear()
      item.tail = tail
  detaileddescription = ''.join(detailednode.itertext()).strip()
  if (len(detaileddescription) > 0):
    print(' {}:'.format(in_blue('Detailed description')))
    print('  {}'.format(detaileddescription))
  print()

#===============================================================================

# No argument, print the full list
if (argument == None):
  pass

# Search name, print the names matching the argument
else:
  name = argument
  if (name in name_list.keys()):
    for (filename,fileref,refid) in (sorted(name_list[name], key=lambda x: x[0])):
      print('  {} in {}'.format(in_red(name),in_red(filename)))
      get_info(fileref,refid)
  else:
    print('Name {} not found'.format(in_red(name)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************

# Compute fitting polynomials for Rys roots and weights.
# The TMax values come from the asymptotic.py script.
# For each order, the range [0,TMax] is split in intervals and for each
# interval a set of fitting polynomials is generated that approximate the
# roots and weights within the interval with the desired accuracy.
# Each interval is expanded as much as possible to keep the fitting accuracy.
# The final part of the output is compatible with the "rysrw" file for Molcas.

from mpmath import mp
from rys_aux import *
import pickle

# Set floating point precision
# Single precision (32 bits): 24-bit significand, 6 decimal places
# Double precision (64 bits): 53-bit significand, 15 decimal places
# Quadruple precision (128 bits): 113-bit significand, 33 decimal places
mp.dps = 33

mRys = 9
fit_degree = 6

order = list(range(1, mRys+1))
only_n = [1, 2, 3, 4, 5, 6, 7, 8, 9]
# Requested accuracy (input)
accuracy = list(map(mp.mpf, ['1e-16', '1e-16', '1e-15', '1e-15', '1e-14', '1e-14', '1e-13', '1e-13', '1e-12']))
# x step (input)
dx = list(map(mp.mpf, ['0.008', '0.007', '0.009', '0.009', '0.012', '0.012', '0.017', '0.016', '0.023']))
fine_step = mp.mpf('0.001')
# x value from which the asymptotic formula gives a relative error in roots and weights lower than the requested accuracy (computed with 0.01 precision)
TMax = list(map(mp.mpf, ['38.80', '46.21', '50.71', '57.03', '60.32', '66.01', '68.69', '74.03', '76.26']))

# This would be with the same accuracy at all orders
#accuracy = [mp.mpf('1e-16') for o in order]
#TMax = list(map(mp.mpf, ['38.80', '46.21', '53.24', '59.64', '65.69', '71.51', '77.16', '82.68', '88.09']))

# Number of data points before reaching "TMax"
nMap = [int(mp.floor(a/b)+1) for a,b in zip(TMax, dx)]
# x values for the data points
xlists = [[] for x in range(len(order))]
for i in range(len(order)):
  for j in range(nMap[i]):
    prev = j*dx[i]
    xlists[i].append(list(frange(prev, (j+1)*dx[i], fine_step)))

# reload the data from previous runs
try:
  with open('tmpfile', 'rb') as f:
    data = pickle.load(f)
except:
  data = {}

for n,acc in zip(order, accuracy):
  if (n not in only_n):
    continue
  xlist = []
  for l in xlists[n-1]:
    xlist.extend(l)

  # Compute roots and weights for different values of x
  # Note that there may be some "duplication" if the end of one interval
  # does not exactly match the beginning of the next (which may occur with frange)
  rtswts = {}
  for x in xlist:
    if (x in rtswts):
      continue
    roots, weights = rysroots(n, x)
    rtswts[x] = [roots, weights]
    print('x = {0}'.format(figs(x)))
    for i in range(len(roots)):
      print('{0}: {1} {2}'.format(i+1, figs(mp.sqrt(roots[i])), figs(weights[i], exp=True)))
  xlist = xlists[n-1]

  # Now proceed with the fitting
  l_min = int(mp.ceil((fit_degree+1)*fine_step/dx[n-1]))+1
  nx = len(xlist)
  num = []
  coeff = []
  center = []
  ini = 0
  print()
  print('{0} points'.format(nx))
  l_prev = 0
  while (ini < nx-1):
    # TODO: bad luck if there are not enough points left to fit
    if (ini+l_min > nx):
      raise
    # l is the number of data points to fit (minimum is l_min)
    # ini is the initial point to include in the fit, so the range is ini to ini+l-1 (ini:ini+l for python)
    l_max = nx-ini
    # start with the same l that worked last time
    l = min(max(l_min, l_prev), l_max)
    while (True):
      print('{0} to {1}'.format(ini, ini+l-1))
      flat_x = [xlist[ini][0]]
      flat_x.extend([x for l in xlist[ini:ini+l] for x in l[1:]])
      x0 = (flat_x[0]+flat_x[-1])/two
      xoff = [x-x0 for x in flat_x]
      mat = []
      for x in xoff:
        mat.append([x**i for i in range(fit_degree, -1, -1)])
      mat = mp.matrix(mat)
      max_error = zero
      cff = []
      # fit for each root (nn) and for roots (i=0) and weights (i=1)
      for nn in range(n):
        for i in range(2):
          val = [rtswts[x][i][nn] for x in flat_x]
          try:
            fit, _ = mp.qr_solve(mat, val)
          except:
            fit = mp.lu_solve(mat, val)
          cff.append(fit.T.tolist()[0])
          error = max([abs((mp.polyval(fit, x)-v)/v) for x,v in zip(xoff, val)])
          max_error = max(max_error, error)
      # if the error is too large, the number of poinst in the fit may have to be reduced
      if (max_error > acc):
        # error if it's already the minimum
        if (l == l_min):
          raise
        # if the last good fit was just one point less, that's the final one
        if (l-l_save == 1):
          l = l_save
          break
        # otherwise reduce to midpoint between current and last good fit
        l_max = l-1
        l = min(int(mp.ceil((l+l_save)/two)), l-1)
      # we have to make sure the saved data correspond to the last good fit
      else:
        cff_save = cff[:]
        x0_save = x0
        l_save = l
        # if l is already at maximum, we're done, otherwise be conservative and step up by 1
        if (l == nx-ini):
          break
        l +=1
    print('fit from {0} to {1}'.format(ini, ini+l-1))
    num.append(l_save)
    ini += l_save
    l_prev = l_save
    coeff.append(cff_save)
    center.append(x0_save)

  # the nesting order for printout is:
  #   roots, weights
  #     0th...6th order coefficients
  #       root 1...n
  #         polynomial 1...npol
  # while in coeff we have:
  #   polynomial 1...npol
  #     root + weight 1...n
  #       6th...0th order coefficients
  coeflist = []
  for i in range(2):
    for j in range(fit_degree, -1, -1):
      for k in range(n):
          for l in range(len(num)):
            coeflist.append(coeff[l][2*k+i][j])
  # map specifying which approximating polynomial to apply for each x
  Map = []
  for i in range(len(num)):
    Map.extend([i+1]*num[i])

  data[n] = {'nx': nx, 'npol': len(num), 'Map': Map, 'center': center, 'coeflist': coeflist}
  # dump the data in a temporary file, so we can run this in several steps
  with open('tmpfile', 'wb') as f:
    pickle.dump(data, f)

# Finally print the data we have available so far
# This is the part that should go in the rysrw file
print()
print(mRys, fit_degree)
printlist(accuracy, 3, f=lambda x: figs(x, n=6, exp=True))
printlist(TMax, 3, f=lambda x: figs(x, n=6))
printlist(dx, 3, f=lambda x: figs(x, n=6))
printlist([i[1]['nx'] for i in sorted(data.items())], 3)
printlist([i[1]['npol'] for i in sorted(data.items())], 3)
for i in sorted(data.items()):
  printlist(i[1]['Map'], 10)
  printlist(i[1]['center'], 6, f=lambda x: figs(x, n=10, exp=True))
for i in sorted(data.items()):
  printlist(i[1]['coeflist'], 4, f=lambda x: figs(x, exp=2, sign=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2018,2019, Ignacio Fdez. Galván                        *
#***********************************************************************

import sys

# These modules should be available in any python installation
standard_modules = [
  '__future__',
  'argparse',
  'ast',
  'base64',
  'binascii',
  'builtins|future.builtins',
  'codecs',
  'contextlib',
  'datetime',
  'decimal',
  'errno',
  'glob',
  'hashlib',
  'io',
  'json',
  'operator',
  'os',
  'os.path',
  'random',
  're',
  'resource',
  'shlex',
  'shutil',
  'signal',
  'stat',
  'subprocess',
  'sys',
  'tempfile',
  'textwrap',
  'threading',
  'types',
  'zlib',
]

# These modules may have to be installed
modules = {
  'pyparsing':[],
  #'setuptools', (only used in pack.py)
}

if (sys.version_info.major < 3):
  modules.update({
    'six':['python_2_unicode_compatible'],
  })

fail = []

try:
  import importlib
except ImportError:
  fail.append('importlib')
else:
  for item in standard_modules + list(modules):
    mods = item.split('|')
    result = []
    for m in mods:
      try:
        tmp = importlib.import_module(m)
        result.append(True)
      except ImportError:
        result.append(False)
    if (not any(result)):
      fail.append(item)
      continue
    if item in modules:
      for i in modules[item]:
        if (i not in dir(tmp)):
          fail.append('{}({})'.format(item, i))

if (fail):
  print(' '.join(fail))

sys.exit(len(fail))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2018, Ignacio Fdez. Galván                             *
#***********************************************************************

import sys
import os
import codecs
import os.path
import re

sys.dont_write_bytecode = True

try:
  docdir = os.path.abspath(sys.argv[1])
except IndexError:
  print('Need doc directory location')
  sys.exit(1)

try:
  datadir = os.path.abspath(sys.argv[2])
except IndexError:
  print('Need data directory location')
  sys.exit(1)

sys.path.append(os.path.join(docdir, 'extensions'))

import xmldoc

xml_re = re.compile(r'(\s*)..\s+xmldoc::\s*')

# Find all ".. xmldoc::" blocks in the source files
xmldocs = {}
for root, subdirs, files in os.walk(os.path.join(docdir, 'source')):
  for fname in files:
    if (not fname.endswith('.rst')):
      continue
    docs = []
    with open(os.path.join(root, fname), 'rb') as f:
      inxml = False
      lines = f.readlines()
      for l in lines:
        try:
          line = l.decode('utf8').rstrip()
        except UnicodeDecodeError:
          continue
        if (inxml):
          match = re.match(r'(\s{{{0}}})'.format(inxml), line)
          if (match or (line == '')):
            try:
              docs[-1].append(line[len(match.group(1)):])
            except AttributeError:
              docs[-1].append('')
          else:
            inxml = False
        if (not inxml):
          match = xml_re.match(line)
          if match:
            inxml = len(match.group(0))
            docs.append([line[inxml:]])
    if (docs):
      xmldocs[f.name] = docs

# Create Help file
with codecs.open(os.path.join(datadir, 'keyword.db'), 'w', 'utf-8') as keywordsfile:
  keywordsfile.write('#This file is generated automatically from the OpenMolcas documentation\n')
  keywordsfile.write('#\n')
  # Sort by filename
  for fname in sorted(xmldocs):
    keywordsfile.write('[%s]\n' % os.path.splitext(os.path.basename(fname))[0].upper())
    for doc in xmldocs[fname]:
      # Reformat the text for the Help file
      text = xmldoc.reformat_Help(doc)
      if (text):
        keywordsfile.write('\n'.join(text)+'\n')

# Create XML file
with codecs.open(os.path.join(datadir, 'keyword.xml'), 'w', 'utf-8') as keywordsfile:
  keywordsfile.write('<!-- This file is generated automatically from the OpenMolcas documentation -->\n')
  keywordsfile.write('<ROOT>\n')
  # Sort by filename
  for fname in sorted(xmldocs):
    for doc in xmldocs[fname]:
      # Reformat the text for XML
      text = xmldoc.reformat_XML(doc)
      if (text):
        keywordsfile.write('\n'.join(text)+'\n')
  keywordsfile.write('</ROOT>\n')




#!/usr/bin/python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2013,2017, Ignacio Fdez. Galván                        *
#               2008,2009, Neil Martinsen-Burrell (FortranFile)        *
#***********************************************************************
# ==============================================================================
# Python script for converting Molcas "grid" files into Gaussian "cube" format
#
# For reading binary grid files, it uses part of the FortranFile library
# (see below)
#
# Last modified: 2017 September 1
#            by: Ignacio Fdez. Galván
# ==============================================================================

import sys, re, os

def usage():
  print ""
  print "Convert a Molcas grid file into a Gaussian cube file."
  print "The input file can be ASCII, binary (non-packed), or in"
  print "Luscus format, the generated output file will be ASCII."
  print ""
  print "USAGE:"
  print "  {0} input_file output_file".format( os.path.basename(__file__) )
  print ""
  print "If there are several grids in the input file,"
  print "the user will be asked which one to convert."
  print ""

# Define symbol list, to get atomic numbers
symbol = [                                              "X",\
  "H",  "HE", "LI", "BE", "B",  "C",  "N",  "O",  "F",  "NE",
  "NA", "MG", "AL", "SI", "P",  "S",  "CL", "AR", "K",  "CA",
  "SC", "TI", "V",  "CR", "MN", "FE", "CO", "NI", "CU", "ZN",
  "GA", "GE", "AS", "SE", "BR", "KR", "RB", "SR", "Y",  "ZR",
  "NB", "MO", "TC", "RU", "RH", "PD", "AG", "CD", "IN", "SN",
  "SB", "TE", "I",  "XE", "CS", "BA", "LA", "CE", "PR", "ND",
  "PM", "SM", "EU", "GD", "TB", "DY", "HO", "ER", "TM", "YB",
  "LU", "HF", "TA", "W",  "RE", "OS", "IR", "PT", "AU", "HG",
  "TL", "PB", "BI", "PO", "AT", "RN", "FR", "RA", "AC", "TH",
  "PA", "U",  "NP", "PU", "AM", "CM", "BK", "CF", "ES", "FM",
  "MD", "NO", "LR", "RF", "DB", "SG", "BH", "HS", "MT", "DS",
  "RG", "CN", "NH", "FL", "MC", "LV", "TS", "OG"
]

# First argument is a grid input file
try:
  grid_input = sys.argv[1]
except IndexError:
  usage()
  sys.exit("** Missing input grid file")

# Second argument is a cube output file
try:
  cube_output = sys.argv[2]
except IndexError:
  usage()
  sys.exit("** Missing output cube file")

#===============================================================================
# THE FOLLOWING IS A MINIMIZED VERSION OF FortranFile
#    ( http://scipy-cookbook.readthedocs.io/items/FortranIO_FortranFile.html )
#
# Copyright 2008, 2009 Neil Martinsen-Burrell
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import struct, numpy

class FortranFile(file):

  def _get_header_length(self):
    return struct.calcsize(self._header_prec)

  _header_length = property(fget=_get_header_length)

  def _set_header_prec(self, prec):
    if prec in 'hilq':
      self._header_prec = prec
    else:
      raise ValueError('Cannot set header precision')

  def _get_header_prec(self):
    return self._header_prec

  HEADER_PREC = property(fset=_set_header_prec,
                         fget=_get_header_prec,
                         doc="Possible header precisions are 'h', 'i', 'l', 'q'"
                        )

  def __init__(self, fname, endian='@', header_prec='i', *args, **kwargs):
    file.__init__(self, fname, *args, **kwargs)
    self.ENDIAN = endian
    self.HEADER_PREC = header_prec

  def _read_exactly(self, num_bytes):
    data = ''
    while True:
      l = len(data)
      if l == num_bytes:
        return data
      else:
        read_data = self.read(num_bytes - l)
      if read_data == '':
        raise IOError('Could not read enough data.'
                      '  Wanted %d bytes, got %d.' % (num_bytes, l))
      data += read_data

  def _read_check(self):
    return struct.unpack(self.ENDIAN+self.HEADER_PREC,
                         self._read_exactly(self._header_length)
                        )[0]

  def readRecord(self):
    l = self._read_check()
    data_str = self._read_exactly(l)
    check_size = self._read_check()
    if check_size != l:
      raise IOError('Error reading record from data file')
    return data_str

  _real_precisions = 'df'

  def readReals(self, prec='f'):
    _numpy_precisions = {'d': numpy.float64,
                         'f': numpy.float32
                        }
    if prec not in self._real_precisions:
      raise ValueError('Not an appropriate precision')
    data_str = self.readRecord()
    num = len(data_str)/struct.calcsize(prec)
    numbers =struct.unpack(self.ENDIAN+str(num)+prec,data_str)
    return numpy.array(numbers, dtype=_numpy_precisions[prec])

# HERE ENDS THE FortranFile PART
#===============================================================================

#===============================================================================
# READ INPUT FILE
#===============================================================================

# Detect LUSCUS, ASCII or binary file
binary = False
luscus = False
file_grid = open(grid_input, "rb")
tell = file_grid.readline()[0]
if (tell != "0"):
  if (tell == " "):
    file_grid.seek(0)
    luscus = True
  else:
    file_grid.close()
    file_grid = FortranFile(grid_input)
    binary = True

grid = []
if (luscus):
  # Hard-coded conversion factor
  angstrom = 0.52917721067
  # Number of atoms, the geometry follows
  line = file_grid.readline()
  natoms = int(line)
  line = file_grid.readline()
  atom = []
  for i in range(natoms):
    tmp = file_grid.readline().split()
    atom.append( { "name": tmp[0], "x": float(tmp[1])/angstrom, "y": float(tmp[2])/angstrom, "z": float(tmp[3])/angstrom } )
  line = file_grid.readline()
  if (line.strip() != '<GRID>'):
    sys.exit("** Error in Luscus format")
  # Number of data per block (per grid)
  line = file_grid.readline()
  match = re.match("\s*N_of_MO=\s*(\d+)\s*N_of_Grids=\s*(\d+)\s*N_of_Points=\s*(\d+)\s*Block_Size=\s*(\d+)\s*N_Blocks=\s*(\d+)\s*Is_cutoff=\s*(\d+)\s*CutOff=\s*(\d+\.\d+)\s*N_P=\s*(\d+)\s*", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  block_size = int(match.group(4))
  # Number of grids
  n_grid = int(match.group(2))
  # Number of grid divisions per dimension
  line = file_grid.readline()
  line = file_grid.readline()
  match = re.match("\s*Net=(.*)", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  tmp = match.group(1).split()
  npt = map(int, tmp)
  # Number of net points is one less in each dimension
  net = map(lambda x: x-1, npt)
  # Origin of the grid (smallest corner)
  line = file_grid.readline()
  match = re.match("\s*Origin=(.*)", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  tmp = match.group(1).split()
  origin = map(float, tmp)
  # The three edges
  line = file_grid.readline()
  match = re.match("\s*Axis_1=(.*)", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  tmp = match.group(1).split()
  axis_1 = map(float, tmp)
  line = file_grid.readline()
  match = re.match("\s*Axis_2=(.*)", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  tmp = match.group(1).split()
  axis_2 = map(float, tmp)
  line = file_grid.readline()
  match = re.match("\s*Axis_3=(.*)", line)
  if (not match):
    sys.exit("** Error in Luscus format")
  tmp = match.group(1).split()
  axis_3 = map(float, tmp)
  # Name of all grids present in the file
  line = file_grid.readline()
  for i in range(n_grid):
    line = file_grid.readline()
    match = re.match("\s*GridName=(.*)", line)
    if (not match):
      sys.exit("** Error in Luscus format")
    grid.append(match.group(1).strip())
else:
  # Read only the header (until the first "Title=" is found)
  while True:
    # Get next line/record in binary or ASCII mode
    if (binary):
      line = file_grid.readRecord()
    else:
      line = file_grid.readline()

    # Number of atoms, the geometry follows
    match = re.match("Natom=\s*(\d+)", line)
    if (match):
      natoms = int(match.group(1))
      atom = []
      for i in range(natoms):
        if (binary):
          tmp = file_grid.readRecord().split()
        else:
          tmp = file_grid.readline().split()
        atom.append( { "name": tmp[0], "x": float(tmp[1]), "y": float(tmp[2]), "z": float(tmp[3]) } )
    # Number of data per block (per grid)
    match = re.match("Block_Size=(.*)", line)
    if (match):
      block_size = int(match.group(1))
    # Number of grid divisions per dimension
    match = re.match("Net=(.*)", line)
    if (match):
      tmp = match.group(1).split()
      net = map(int, tmp)
      # Number of points is one more in each dimension
      npt = map(lambda x: x+1, net)
    # Origin of the grid (smallest corner)
    match = re.match("Origin=(.*)", line)
    if (match):
      tmp = match.group(1).split()
      origin = map(float, tmp)
    # The three edges
    match = re.match("Axis_1=(.*)", line)
    if (match):
      tmp = match.group(1).split()
      axis_1 = map(float, tmp)
    match = re.match("Axis_2=(.*)", line)
    if (match):
      tmp = match.group(1).split()
      axis_2 = map(float, tmp)
    match = re.match("Axis_3=(.*)", line)
    if (match):
      tmp = match.group(1).split()
      axis_3 = map(float, tmp)
    # Name of all grids present in the file
    match = re.match("GridName=(.*)", line)
    if (match):
      grid.append(match.group(1).strip())
    # The first grid data starts, stop reading header
    match = re.match("Title=(.*)", line)
    if (match):
      break

#---------------------------------------
# Select a grid from the input file
#---------------------------------------

# Print a menu showing the grids found in the file
if (len(grid) == 1):
  ng = 1
if (len(grid) < 1):
  sys.exit("** No grids found in input file")
if (len(grid) > 1):
  print "Grids in input file:"
  for i in range(len(grid)):
    print "{0:3}: {1}".format( i+1, grid[i] )
  # Read the user selection
  try:
    ng = int(raw_input("Which one to convert? "))
  except ValueError:
    sys.exit("** Sorry, I don't understand")
  if (ng < 1 or ng > len(grid)):
    sys.exit("** Sorry, no such grid")

# Total number of grid points to read
np = npt[0]*npt[1]*npt[2]
data = []

if (luscus):
  loc = 0
  while True:
    line = file_grid.readline()
    if (re.match("\s*<DENSITY>", line)):
      loc = file_grid.tell()
      break
  if (loc == 0):
    sys.exit("** Error in Luscus format")
  l = struct.calcsize('d')
  # number of points read
  npr = 0
  # read the points by blocks
  while (npr < np):
    # number of points in the block
    npb = min(np-npr, block_size)
    # skip previous grids
    file_grid.seek((ng-1)*npb*l, 1)
    # read the block for this grid
    data.extend( list(struct.unpack(npb*'d', file_grid.read(npb*l))) )
    npr += npb
    # skip following grids
    file_grid.seek((n_grid-ng)*npb*l, 1)
else:
  # The data for each grid is written across several blocks separated by "Title="
  i = 0
  j = 0
  file_grid.seek(0)
  if (binary):
    line = file_grid.readRecord()
  else:
    line = file_grid.readline()
  while True:
    # When a new block starts, step the grid count by 1
    # (and start again when the last grid is reached)
    match = re.match("Title=(.*)", line)
    if (match):
      i = (i % len(grid))+1
    # Only read the data if this is the grid we want
    if (i == ng):
      if (binary):
        # Assume we are reading double precision
        line = file_grid.readReals(prec='d')
        data.extend(line)
        j += len(line)
      else:
        # Each block is at most of block_size length (the last one is shorter)
        for k in range(min(np-j, block_size)):
          data.append( float(file_grid.readline().split()[0]) )
        j += k+1
    if (j >= np): break
    if (binary):
      line = file_grid.readRecord()
    else:
      line = file_grid.readline()

file_grid.close()

#===============================================================================
# WRITE OUTPUT FILE
#===============================================================================

# Set the delta values in each dimension
dx = map( lambda x: x/float(net[0]), axis_1 )
dy = map( lambda x: x/float(net[1]), axis_2 )
dz = map( lambda x: x/float(net[2]), axis_3 )
# Assign an atomic number for each atom
# (from symbol, which is assumed to be the first 1 or 2 letters)
for i in atom:
  match = re.match("([A-Za-z]{1,2})", i["name"])
  i["number"] = symbol.index(match.group(1).upper())

file_cube = open(cube_output, "w")

# Write the header
file_cube.write("File converted from MOLCAS grid format\n")
file_cube.write("Title = {0}\n".format( grid[ng-1] ))
# Grid properties
file_cube.write("{0:5}{1:12.6f}{2:12.6f}{3:12.6f}\n".format( natoms, *origin ))
file_cube.write("{0:5}{1:12.6f}{2:12.6f}{3:12.6f}\n".format( npt[0], *dx ))
file_cube.write("{0:5}{1:12.6f}{2:12.6f}{3:12.6f}\n".format( npt[1], *dy ))
file_cube.write("{0:5}{1:12.6f}{2:12.6f}{3:12.6f}\n".format( npt[2], *dz ))
# Molecular geometry
for i in atom:
  file_cube.write("{0:5}{1:12.6f}{2:12.6f}{3:12.6f}{4:12.6f}\n".format( i["number"], 0.0, i["x"], i["y"], i["z"] ))

# Write the grid data, nd values in each line, new line every npt[2] elements
k = 0
nd = 6 # default 6
for i in range(npt[0]):
  for j in range(npt[1]):
    # kk is the final element of this block (i,j pair)
    kk = k+npt[2]
    # end is the final relative element of this (previous) line
    end = 0
    while (k+end < kk):
      # next element to print is k+end
      k += end
      # final element of this line is either kk or k+nd
      end = min(kk-k, nd)
      # write elements from k to k+end
      file_cube.write("".join([ "{0:13.5E}".format(num) for num in data[k:k+end] ] )+"\n")
    # update for the final line of the block
    k += end

file_cube.close()



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#***********************************************************************

#***********************************************************************
# This file is based on code posted by J.F. Sebastian on stackoverflow *
# (https://stackoverflow.com/a/4985080)                                *
# and licensed under CC-BY-SA 3.0                                      *
# (https://creativecommons.org/licenses/by-sa/3.0)                     *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
from subprocess import Popen, PIPE
from threading import Thread

def tee(infile, *files):
  '''Print "infile" to "files" in a separate thread.'''
  def fanout(infile, *files):
    # Write each input line in all the files,
    # Try as a byte string first, then as utf8
    for line in iter(infile.readline, b''):
      for f in files:
        try:
          f.write(line)
        except TypeError:
          f.write(line.decode('utf-8', 'replace'))
    infile.close()
  t = Thread(target=fanout, args=(infile,)+files)
  t.daemon = True
  t.start()
  return t

def teed_call(cmd_args, **kwargs):    
  stdout, stderr = [kwargs.pop(s, None) for s in ['stdout', 'stderr']]
  no_tee = kwargs.pop('no_tee', False)
  if (no_tee):
    p = Popen(cmd_args, stdout=sys.stdout, stderr=sys.stderr, **kwargs)
  else:
    p = Popen(cmd_args,
            stdout=PIPE if stdout is not None else None,
            stderr=PIPE if stderr is not None else None,
            **kwargs)
    threads = []
    if stdout is not None: threads.append(tee(p.stdout, stdout, sys.stdout))
    if stderr is not None: threads.append(tee(p.stderr, stderr, sys.stderr))
    for t in threads:
      t.join()
  return p.wait()



#! /usr/bin/env python

ATOMIC_NUMBERS={
"H": 1,
"He": 2,
"Li": 3,
"Be": 4,
"B": 5,
"C": 6,
"N": 7,
"O": 8,
"F": 9,
"Ne": 10,
"Na": 11,
"Mg": 12,
"Al": 13,
"Si": 14,
"P": 15,
"S": 16,
"Cl": 17,
"Ar": 18,
"K": 19,
"Ca": 20,
"Sc": 21,
"Ti": 22,
"V": 23,
"Cr": 24,
"Mn": 25,
"Fe": 26,
"Co": 27,
"Ni": 28,
"Cu": 29,
"Zn": 30,
"Ga": 31,
"Ge": 32,
"As": 33,
"Se": 34,
"Br": 35,
"Kr": 36,
"Rb": 37,
"Sr": 38,
"Y": 39,
"Zr": 40,
"Nb": 41,
"Mo": 42,
"Tc": 43,
"Ru": 44,
"Rh": 45,
"Pd": 46,
"Ag": 47,
"Cd": 48,
"In": 49,
"Sn": 50,
"Sb": 51,
"Te": 52,
"I": 53,
"Xe": 54,
"Cs": 55,
"Ba": 56,
"La": 57,
"Hf": 72,
"Ta": 73,
"W": 74,
"Re": 75,
"Os": 76,
"Ir": 77,
"Pt": 78,
"Au": 79,
"Hg": 80,
"Tl": 81,
"Pb": 82,
"Bi": 83,
"Po": 84,
"At": 85,
"Rn": 86,
"Fr": 87,
"Ra": 88,
"Ac": 89,
"Ku": 104,
"Ha": 105
}

class Parser:

   def __init__(self,infile_name):
      self.infile_name=infile_name
      try:
         i=open(self.infile_name,'r')
      except IOError:
         print "No such file or directory:",self.infile_name
         import sys
         sys.exit()
      input=i.readlines()
      i.close()
      self.molecules=self._parse(input)

   def _parse(self,input):
      molecules={}
      count=0
      for line in input:
         if line[0] == '#':
            if count > 0:
               molecules[name]=data
            name=line[1:].split()[0]
            if name in molecules.keys():
               print name,"used more than once"
               import sys
               sys.exit(1)
            data=[]
            count+=1
         else:
            l=line.split()
            a=l[3]
            atom=a[0].upper()+a[1:]
            d=[atom,float(l[0]),float(l[1]),float(l[2])]
            data.append(d)
      if count > 0:
         molecules[name]=data
      return molecules

   def Molecules(self):
      return self.molecules.keys()

   def NumberOfMolecules(self):
      return len(self.Molecules())

   def Atoms(self,molecule):
      return [x[0] for x in self.molecules[molecule]]

   def toXYZ(self,molecule,comment="",conversion_factor=0.52917720859):
      c=conversion_factor
      s=""
      if molecule in self.Molecules():
         atoms=self.molecules[molecule]
         s+=str(len(atoms))+"\n"
         s+=molecule+" "+comment+"\n"
         for atom in atoms:
            s+="%-6s  %20.14f  %20.14f  %20.14f \n" %(atom[0],c*atom[1],c*atom[2],c*atom[3])
      return s

   def AlltoXYZ(self,comment=""):
      s=""
      for molecule in self.Molecules():
         s+=self.toXYZ(molecule,comment=comment)
      return s

   def toSewardInput(self,molecule,basis="$Basis.....",unit="/Angstrom",conversion_factor=0.52917720859):
      c=conversion_factor
      s=""
      if molecule in self.Molecules():
         s+=" &SEWARD &END\n"
         s+="\nTitle\n"
         s+=" "+molecule+"\n\n"
         labels={}
         atoms=self.molecules[molecule]
         for atom in atoms:
            if atom[0] in labels.keys():
               labels[atom[0]]+=1
            else:
               labels[atom[0]]=1
            mylabel=atom[0]+"_"+str(labels[atom[0]])
            s+="\nBasis Set \n"
            s+=atom[0]+"."+basis+"\n"
            s+="%-6s  %20.14f  %20.14f  %20.14f %-10s \n" %(mylabel,c*atom[1],c*atom[2],c*atom[3],unit)
            s+="End Of Basis\n"
         s+="\nEnd Of Input"
      return s

   def toBlockedSewardInput(self,molecule,basis="$Basis",unit="/Angstrom",conversion_factor=0.52917720859):
      c=conversion_factor
      s=""
      if molecule in self.Molecules():
         s+=" &SEWARD &END\n"
         s+="\nTitle\n"
         s+=" "+molecule+"\n\n"
         labels={}
         atoms=self.molecules[molecule]
         for atom in atoms:
            if atom[0] in labels.keys():
               labels[atom[0]]+=1
            else:
               labels[atom[0]]=1
         for label in labels.keys():
            s+="\nBasis Set\n"
            s+=label+"."+basis+"\n"
            i=0
            for atom in atoms:
               if atom[0] == label:
                  i+=1
                  if labels[label] == 1:
                     mylabel=atom[0]
                  else:
                     mylabel=atom[0]+"_"+str(i)
                  s+="%-6s  %20.14f  %20.14f  %20.14f %-10s \n" %(mylabel,c*atom[1],c*atom[2],c*atom[3],unit)
            s+="End Of Basis\n"
         s+="\nEnd Of Input"
      return s

def SCFInput(*args):
    s="\n &SCF &END\n"
    for arg in args:
        s+=str(arg)+"\n"
    s+="End Of Input"
    return s

def addKeywordsToInput(input,*args):
    s=""
    for line in input.split("\n"):
       if line == "End Of Input":
          for arg in args:
             s+=str(arg)+"\n"
          s+="\n"
       s+=line+"\n"
    return s


if __name__=="__main__":
   import sys
   inp=sys.argv[1:]
   if len(inp)==0:
      print "No input file specified"
      sys.exit()
   # parse input file
   data=Parser(inp[0])
   # get list of molecules
   molecules=data.Molecules()
   # sort the list alphabetically (in place)
   molecules.sort()
   # for each molecule:
   for molecule in molecules:
      # get a seward input file (a string), assuming bohr as unit in input file
      input=data.toBlockedSewardInput(molecule,basis="$Basis",unit="/bohr",conversion_factor=1.0e0)
      # add CD keywords to seward input
      input=addKeywordsToInput(input,"$Seward_Options")
      # Count number of electrons in neutral system
      Ne=0
      for atom in data.Atoms(molecule):
         Ne+=ATOMIC_NUMBERS[atom]
      # add an SCF input section (set charge if needed)
      if molecule[-1] == "-":
         # add 1 to number of electrons
         Ne+=1
         # UHF for odd number of electrons, else RHF
         if Ne%2==1:
            model="UHF"
            input+=SCFInput("UHF","Charge"," -1","$SCF_Options")
         else:
            model="RHF"
            input+=SCFInput("Charge"," -1","$SCF_Options")
      else:
         # UHF for odd number of electrons, else RHF
         if Ne%2==1:
            model="UHF"
            input+=SCFInput("UHF","$SCF_Options")
         else:
            model="RHF"
            input+=SCFInput("$SCF_Options")
      # write input to file molecule.inp
      file="input/"+molecule+".input"
      f=open(file,"w")
      print >>f, input
      print molecule,"\t",file,"\t",model
      f.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015,2017,2018, Ignacio Fdez. Galván                   *
#***********************************************************************

'''
This exports pymolcas to a single-file python script that can be copied around.
It still uses python and it's trivial to recover the original files (although
without comments). Obfuscation is not the goal here, just getting something
that's easy to run, move and distribute.
'''

from __future__ import (unicode_literals, division, absolute_import, print_function)
try:
  from builtins import (bytes, str)
except:
  from future.builtins import (bytes, str)
from io import open

import sys, zlib, base64, os, stat
sys.dont_write_bytecode = True

files = ['tee', 'molcas_aux', 'emil_grammar', 'simpleeval', 'abstract_flow', 'emil_parse', 'python_parse', 'check_test', 'validate', 'write_molcasrc', 'molcas_wrapper', 'pymolcas']
try:
  exe_name = sys.argv[1]
except:
  exe_name = 'pymolcas'
obfuscate = False # only obfuscate the "wrapper", not the modules' code (yet)
try:
  import pyminifier.minification as pm
  compact = True
except ImportError:
  compact = False
compress_and_b64 = True

def wrap(text, width):
  lines = []
  l = 0
  while (l < len(text)):
    r = min(l+width,len(text)+1)
    lines.append(text[l:r])
    l = r
  return '\n'.join(lines)

def minify(string):
  string = pm.remove_comments_and_docstrings(string)
  string = pm.multiline_indicator.sub('', string)
  string = pm.fix_empty_methods(string)
  string = pm.join_multiline_pairs(string)
  string = pm.join_multiline_pairs(string, '[]')
  string = pm.join_multiline_pairs(string, '{}')
  string = pm.remove_blank_lines(string)
  string = pm.reduce_operators(string)
  string = pm.dedent(string)
  return string

if (compress_and_b64):
  code = 'import zlib,base64;exec(zlib.decompress(base64.b64decode(bytes(m[1],\\\'ascii\\\'))),module.__dict__);del zlib,base64'
else:
  code = 'exec(m[1],module.__dict__)'

hexcode = ''.join('{:02x}'.format(c) for c in bytes(code, 'ascii'))

from pymolcas import warning

mods_name = 'M' if obfuscate else 'modules'

failed = False

with open(exe_name, 'w', encoding='utf-8') as f:
  interpreter = sys.executable
  if (interpreter is None):
    interpreter = '/usr/bin/env python'
  f.write('''#!{0}
# -*- coding: utf-8 -*-
#{1}
{2} = [

'''.format(interpreter, warning.replace('\n','\n#'), mods_name))

  try:

    for i in files:
      filename = '{0}.py'.format(i)
      with open(filename, 'r', encoding='utf-8') as fin:
        content = fin.read()
      if (compact):
        try:
          #python3
          content = minify(content)
        except UnicodeEncodeError:
          #python2
          content = minify(bytes(content, 'utf-8'))
      if (compress_and_b64):
        content = str(base64.b64encode(zlib.compress(bytes(content, 'utf-8'))), 'utf-8')
        content = wrap(content, 120)
        fmt = "  ['{0}', '''\n{1}\n'''],\n\n"
      else:
        fmt = "  ['{0}', \n{1}\n],\n\n"
      content = bytes(content, 'utf-8')
      if (obfuscate):
        name_i = ''.join('{:02x}'.format(c) for c in bytes(i, 'ascii'))
      else:
        name_i = i
      f.write(fmt.format(name_i, content.decode('utf-8')))

    f.write(']\n')

    if (obfuscate):
      f.write('''
checksum = '\''
{0}
'\''.replace('\\n', '')
'''.format(wrap(hexcode,120)))
      mod_name_code = 'binascii.unhexlify(m[0]).decode(\'ascii\')'
      mod_code = 'binascii.unhexlify(checksum)'
    else:
      mod_name_code = 'm[0]'
      mod_code = '\'{0}\''.format(code)

    f.write('''
import sys, shutil, os.path, types, binascii
try:
  from builtins import bytes
except ImportError:
  from future.builtins import bytes
for m in {0}:
  x = {1}
  module = types.ModuleType(x)
  exec({2})
  sys.modules[x] = module
del {0}, types, binascii

from pymolcas import main
from molcas_aux import which

f = which(sys.argv[0])
if (f is None):
  f = sys.argv[0]
sys.exit(main(os.path.realpath(f)))
'''.format(mods_name, mod_name_code, mod_code))

  except Exception as e:
    print(str(e))
    failed = True

if (failed):
  os.remove(exe_name)
  sys.exit(1)
else:
  st = os.stat(exe_name)
  os.chmod(exe_name, st.st_mode | stat.S_IEXEC)

sys.exit(0)



# -*- coding: utf-8 -*-
#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Ignacio Fdez. Galván                             *
#***********************************************************************

import re
import textwrap
from xml.etree import ElementTree as ET

# Return the number of leading spaces and the "item label" format, if any
# The "item label" is the first word if it is an asterisk or a single
# letter or number followed by . or )
#
def _indentlabel(line):
  match = re.match(r'(\s*)(\S*)?', line)
  indent = len(match.group(1))
  label = match.group(2)
  match = re.match(r'[a-zA-Z0-9][.)]$', label)
  if label == '*':
    pass
  elif match:
    if re.match(r'[a-z]', label):
      label = 'a' + label[1]
    elif re.match(r'[A-Z]', label):
      label = 'A' + label[1]
    elif re.match(r'[0-9]', label):
      label = '1' + label[1]
  else:
    match = re.match(r'(\s*)(\S+\s+-+)', line)
    if match:
      label = len(match.group(2))*'_'
    else:
      label = ''
  return indent, label

# Return the number of spaces that should be used for "hanging indent",
# i.e., the number of leading spaces (plus the length of the item label plus one)
#
def _hangindent(label):
  if label[2]:
    return label[1] + len(label[2]) + 1
  else:
    return label[1]

# Finish a paragraph-level block, by assigning the current buffer text
#
def _finish_paragraph(element, paragraph):
  if element is not None:
    element.text = '\n'.join(paragraph).rstrip()
  elif paragraph:
    raise Exception
  return None, []

# Append an element updating the parent map
#
def _append_element(element, parent, parent_map):
  parent.append(element)
  parent_map[element] = parent

# Parse a help text, returning an HTML-like tree
#
def parse_help_text(text):
  el = None
  par = []
  level = [(None, -1, '')]
  newpar = False
  doc = ET.Element('root')
  parent_map = {doc: None}

  for line in text.split('\n'):
    ind, lab = _indentlabel(line)
    blank = not line.strip()
    # A pre-formatted block continues until a line with less indent is found
    if el is not None and el.tag == 'pre' and (ind >= level[-1][1] or blank):
      par.append(line)
      continue
    # A paragraph-block is finished by a blank line
    # or a line that does not match the current hanging indent level
    if blank or ind != _hangindent(level[-1]):
      el, par = _finish_paragraph(el, par)
      newpar = True
    # If a new paragraph starts, find out which type and level
    if not blank and newpar:
      newpar = False
      # Start from innermost level and go outwards, until a match is found
      for i in range(len(level)):
        parent = level[-1][0]
        if parent is None:
          parent = doc
        # Normal paragraph at the current level
        if not lab and ind == _hangindent(level[-1]):
          break
        # List item at the current level
        elif (ind, lab) == level[-1][1:3]:
          # Create a new <li> element in the parent list, and update the current parent
          li = ET.Element('li')
          _append_element(li, parent_map[parent], parent_map)
          parent = li
          level[-1] = (parent,) + level[-1][1:]
          break
        # New list, or more indent
        elif (lab and ind == level[-1][1]) or ind > level[-1][1]:
          # New list: create new <ul> and <li> elements
          if lab:
            ul = ET.Element('ul')
            ul.attrib['label'] = lab
            _append_element(ul, parent, parent_map)
            parent = ul
            li = ET.Element('li')
            _append_element(li, parent, parent_map)
            parent = li
          # Indented paragraph: child of the last added element
          elif len(parent) > 0:
            parent = parent[-1]
          level.append((parent, ind, lab))
          break
        del level[-1]
      # A child of <p> is actually a sibling <pre>
      if parent.tag == 'p':
        parent = parent_map[parent]
        el = ET.Element('pre')
      # Everything else is a <p> (possibly inside <li>)
      else:
        el = ET.Element('p')
      _append_element(el, parent, parent_map)
    # Whatever the case, add the current line to the buffer
    if not blank:
      par.append(line)
  # Finish the last element
  el, par = _finish_paragraph(el, par)

  return doc

# Format an HTML element as plain text
#
def _format_element(element, parent_map, textwidth, indent):
  out = ''
  # Increase indent in lists by multiples of 2
  if element.tag in ['ol', 'ul']:
    indent += 2
  # Increase indent of preformatted texts by 4
  elif element.tag == 'pre':
    indent += 4
  # If there are child elements, call recursively
  if len(element) > 0:
    for el in element:
      out += _format_element(el, parent_map, textwidth, indent)
  # If there are no child elements, format the inner text
  else:
    # Pre-formatted text: Verbatim lines (replacing common indentation)
    if element.tag == 'pre':
      for line in textwrap.dedent(element.text).split('\n'):
        if line:
          out += indent*' ' + line + '\n'
        else:
          out += '\n'
    # Other elements, use textwrap
    elif element.text:
      tw = textwrap.TextWrapper(break_long_words=False, break_on_hyphens=False, width=textwidth)
      tw.initial_indent = indent*' '
      tw.subsequent_indent = tw.initial_indent
      # List items: Hanging indent in first child, extra indent in the rest
      parent = parent_map[element]
      text = element.text.lstrip()
      lab = ''
      if parent.tag == 'li':
        hang = len(parent_map[parent].attrib['label'])
        tw.subsequent_indent += (hang+1)*' '
        if element is parent[0]:
          # Split the label, so the spaces here are not altered
          lab = text[0:hang] + ' '
          text = text[hang:]
        else:
          tw.initial_indent += hang*' '
      text = lab + ' '.join(text.split())
      out += tw.fill(text) + '\n'
    out += '\n'

  return out

# Return a plain-text-formatted text from an HTML-like tree
#
def format_help_text(doc, textwidth=80):
  parent_map = {e:p for p in doc.iter() for e in p}
  return _format_element(doc, parent_map, textwidth, 0).strip()



# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2015-2017, Ignacio Fdez. Galván                        *
#***********************************************************************

from __future__ import (unicode_literals, division, absolute_import, print_function)

from re import sub
from pyparsing import *

# helpful methods (python 2/3 compatibility)
def chomp(s):
  return s[:-1] if s.endswith('\n') else s

def chompAction(s, l, t):
  try:
    return list(map(lambda s: chomp(unicode(s)), t))
  except NameError:
    return list(map(chomp, t))

def removeEMILEnd(s):
  return sub(r'\s*<*\s*$','',s)

def removeEMILEndAction(s, l, t):
  try:
    return list(map(lambda s: removeEMILEnd(unicode(s)), t))
  except NameError:
    return list(map(removeEMILEnd, t))

################################################################################
# Attempt to specify a grammar for EMIL (extended Molcas input language)
################################################################################

# do not skip newlines when parsing
ParserElement.setDefaultWhitespaceChars(' \t')

# define a class that grabs the rest of the line, dropping the newline character
NL = Suppress( LineEnd() )
restOfLineNL = restOfLine + NL

# empty lines (between programs/commands)
EmptyLines = OneOrMore( NL )

# the same, but removing the optional "<" and spaces at the end of the line
restOfEMILLine = restOfLine.setParseAction(removeEMILEndAction)
restOfEMILLineNL = restOfEMILLine + NL

# and another that stops at a semicolon
RestComment = Suppress( Literal('//') + restOfLineNL )
restOfLineSemiColon = SkipTo( ( NL | RestComment ) | Suppress(Literal(';')), include=True )

# this seems to catch lines starting with spaces more reliably
#LineStartWS = Suppress( LineStart().leaveWhitespace() )
LineStartWS = Suppress( ZeroOrMore(White(' ')) )

# comments (line and block, and trailing newlines) are removed
LineComment = ( Literal('*') | Literal('//') ) + restOfLineNL
BlockComment = Literal('/*') + SkipTo( StringEnd() | Literal('*/'), include=True ) + Optional(NL)
TrailingNL = OneOrMore( NL ) + StringEnd()
Comment = Suppress( TrailingNL | BlockComment | LineComment )

# convenience alias
AnyLine = restOfLineSemiColon

# one-line EMIL commands start with one or more '>' and extend to the end of the line
EMILMark = (OneOrMore('>') + Optional(White())).setParseAction(replaceWith('>'))
EMILCommand = LineStartWS + Group(EMILMark + restOfEMILLineNL)

# the FILE command requires a closing line
EMILEOF = EMILMark + CaselessLiteral('EOF') + restOfEMILLineNL
EMILFileMark = ( originalTextFor( CaselessLiteral('FILE') + restOfLine ) + NL ).setParseAction(removeEMILEndAction)
EMILFile = LineStartWS + Group(EMILMark + EMILFileMark + SkipTo(EMILEOF).leaveWhitespace().setParseAction(chompAction)) + Suppress(EMILEOF)

# at the moment only FILE is a block command
EMILBlock = EMILFile

EMIL = EMILBlock | EMILCommand

# modules start with their name (preceded with '&') and contain every line
# until another module starts or an EMIL command is found
ModuleName = LineStartWS + Literal('&') + Word( alphanums + '_', min=2 ) + Suppress(restOfLineSemiColon)
ModuleEnd = StringEnd() | EMIL | ModuleName
Module = Group( ModuleName + ZeroOrMore( ~ModuleEnd + ( Comment | AnyLine ) ) )

# the input file should contain only modules, commands and comments
EMIL_Grammar = ZeroOrMore( EMIL | Module | Comment | EmptyLines )

################################################################################



#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2020, Oskar Weser                                      *
#***********************************************************************

from numpy import eye
from scipy.linalg import svd, det

def get_kabsch_transformation(Q, P, force_rotation=False):
    """Calculate the optimal orthogonal transformation from ``P`` onto ``Q``.

    If force_rotation is true, the transformation
    is from the special orthogonal group.
    """
    # Naming of variables follows the wikipedia article:
    # http://en.wikipedia.org/wiki/Kabsch_algorithm
    A = P.T @ Q
    # One can't initialize an array over its transposed. :-(
    U, S, V = svd(A)
    V = V.T
    if force_rotation:
        sign = eye(W)
        # NOTE: W and V are unitary. Up to float noise we know that:
        #   det(W @ V.T) in {-1, +1}
        sign[-1, -1] = 1.0 if det(V @ U.T) > 0.0 else -1.0
        return V @ sign @ U.T
    else:
        return V @ U.T



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************

# Generate alpha and beta recursion coefficients for the Rys polynomials
# Output compatible with the "abdata" file for Molcas

from mpmath import mp
from rys_aux import *

# Set floating point precision
# Single precision (32 bits): 24-bit significand, 6 decimal places
# Double precision (64 bits): 53-bit significand, 15 decimal places
# Quadruple precision (128 bits): 113-bit significand, 33 decimal places
mp.dps = 33

tabini = -2
tabend = 678
preci = mp.mpf('1e-20')
terms = 20

# Print header
print('''Data tables from genab.py
=========================

Each table gives the recursion parameters ALPHA(K), BETA(K) and the
zeroth order polynomial P0 needed to generate Rys polynomials up to
order MAXDEG for different values of the continuous parameter T.
NTAB1 and NTAB2 are the first and last table index. PRECIS is the
accuracy to which the values were computed. For each table with index
ITAB, the corresponding T value is:

  T = (sqrt(72900-(260-ITAB)*ITAB)+ITAB-270)/10

and the inverse formula for later interpolation:

  ITAB = 5*T+200*T/(14+T)

NTAB1, NTAB2, MAXDEG:
{0} {1} {2}
PRECIS
{3}'''.format(tabini, tabend, terms-1, preci))

# Compute and print alpha and beta values for different x values
for i in range(tabini, tabend+1):
  # Get the x value from the index
  x = mp.mpf(mp.sqrt(72900-(260-i)*i)+i-270)/10
  alpha, beta = rysab(terms, x, preci=preci)
  # Transform to the Molcas convention
  beta = [mp.sqrt(b) for b in beta]
  p0 = one/beta[0]
  beta[0] = zero

  # Print the results
  print('TAB POINT NR., T VALUE, P0 VALUE:')
  print('{0} {1} {2}'.format(i, figs(x), figs(p0)))
  print('ALPHA ARRAY:')
  printlist(alpha, 4, f=figs)
  print('BETA ARRAY:')
  printlist(beta, 4, f=figs)



#!/usr/bin/env python3
"""Script to automatically generate grid files from Orbfiles.
"""

import os
from os import remove
from os.path import join, basename, splitext, relpath

import subprocess
from subprocess import run
import click


def create_module_input(module_name, **kwargs):
    indentstep = 2
    indentation = indentstep * ' '

    out = f'&{module_name.upper()}\n'
    for keyword, value in kwargs.items():
        out += f'{1 * indentation}{keyword}\n'
        if value:
            out += f'{2 * indentation}{value}\n'
    out += 'End of Input\n'
    return out


def emil_cmd(cmd, *args):
    return f'> {cmd} ' + ' '.join(args) + '\n'


def get_currdir_path(path):
    return join('$CurrDir', relpath(path, os.getcwd()))


def create_input(coord, basis, inporb, density='sparse', group=None,
                 select=None, total=False):
    inp = create_module_input(
        'GATEWAY',
        coord=get_currdir_path(coord),
        basis=basis, group='full' if group is None else group)

    inp += emil_cmd('COPY', get_currdir_path(inporb), 'INPORB')

    gridit = {'NOLUSCUS': '', 'ASCII': '', density: ''}
    if select:
        gridit['select'] = select
    if total:
        gridit['total'] = ''

    inp += create_module_input('GRIDIT', **gridit)
    return inp


def start_calc(input_str, inp_path='make_grid.inp', molcas_exe='pymolcas'):

    with open(inp_path, 'w') as f:
        f.write(input_str)

    run([molcas_exe, '-f', basename(inp_path)], stdout=subprocess.PIPE,
        check=True)


def get_project_name(inporb):
    return splitext(basename(inporb))[0]


def cleanup(project):
    remove(f'{project}.inp')
    remove(f'{project}.log')
    remove(f'{project}.status')
    remove(f'{project}.err')
    remove('xmldump')


@click.command(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--coord', '-c', required=True, type=str,
              help='Filepath to xyz-file.')
@click.option('--basis', '-b', required=True, type=str,
              help='Basisset to use in Molcas nomenclature.')
@click.option('--inporb', '-i', required=True, type=str,
              help='Filepath to Molcas orbital file.')
@click.option('--density', '-d', default='sparse', type=str, show_default=True,
              help='Gridit keyword. One of {sparse, normal, dense}.')
@click.option('--group', '-g', default='full', type=str, show_default=True,
              help='Symmetry group in Molcas nomenclature.')
@click.option('--select', '-s', default=None, type=str, show_default=True,
              help=('Gridit keyword to select certain orbitals. '
                    'The default depends on the Molcas implementation. '
                    'Usually the active space is used.'))
@click.option('--total/--no-total', '-t',default=False,
              help='Total density computed from contributions of all orbitals.')
@click.option('--project', '-p', default='make_grid', type=str,
              show_default=True,
              help=('Prefix for the grid calculations. '
                    'The gridfile will be named ${project}.grid.'))
@click.option('--molcas_exe', '-m', default='pymolcas', type=str,
              show_default=True,
              help='Filepath of the Molcas driver script.')
@click.option('--clean/--no-clean', default=True,
              help='Clean temporary files after calculation.')
def main(coord, basis, inporb, density, group, select, total, project, molcas_exe, clean):
    start_calc(
        create_input(coord, basis, inporb, density, group, select, total),
        f'{project}.inp', molcas_exe)
    if clean:
        cleanup(project)



if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#***********************************************************************
# This file is part of OpenMolcas.                                     *
#                                                                      *
# OpenMolcas is free software; you can redistribute it and/or modify   *
# it under the terms of the GNU Lesser General Public License, v. 2.1. *
# OpenMolcas is distributed in the hope that it will be useful, but it *
# is provided "as is" and without any express or implied warranties.   *
# For more details see the full text of the license in the file        *
# LICENSE or in <http://www.gnu.org/licenses/>.                        *
#                                                                      *
# Copyright (C) 2017, Ignacio Fdez. Galván                             *
#***********************************************************************

# Compute the minimum parameter x for which the roots and weights of Rys polynomials
# can be obtained with the desired accuracy from the asymptotic formula.
# The list of orders for Rys polynomials is specified in "order", the desired
# accuracy in "accuracy", and the precision for the result is the last element in "step".

from mpmath import mp
from rys_aux import *

# Set floating point precision
# Single precision (32 bits): 24-bit significand, 6 decimal places
# Double precision (64 bits): 53-bit significand, 15 decimal places
# Quadruple precision (128 bits): 113-bit significand, 33 decimal places
mp.dps = 33

# Order of the Rys polynomials (1 to 9)
order = list(range(1, 10))
# Requested accuracy (input)
# Higher orders will typically have a smaller contribution, so the accuracy can be lower
accuracy = list(map(mp.mpf, ['1e-16', '1e-16', '1e-15', '1e-15', '1e-14', '1e-14', '1e-13', '1e-13', '1e-12']))
# Precision for the limits (several passes)
step = list(map(mp.mpf, ['10', '1', '0.01']))

# Uncomment to get same accuracy all along
#accuracy = [mp.mpf('1e-16') for o in order]

# Do several passes, increasing x by the specified step and stopping when
# the error is below the requested accuracy
# Start from zero and then use the result from the previous pass
init = [mp.mpf(0) for i in order]
for s in step:
  asymp = []
  for n,x0,acc in zip(order, init, accuracy):

    print('Starting n={0}'.format(n))

    # Generate roots and weights for the asymptotic formula (Hermite)
    #   alpha(i) = 0
    #   beta(0)  = sqrt(pi)
    #   beta(i)  = i/2
    alpha_herm = [zero for i in range(2*n)]
    beta_herm = [mp.sqrt(pi)]
    for i in range(1, 2*n):
      beta_herm.append(i/two)
    roots_herm, weights_herm = gauss(alpha_herm, beta_herm, mp.mpf('1e-20'))

    i = 0
    while True:
      x = x0+i*s
      i += 1

      # Generate roots and weights of Rys polynomial of order n for parameter x
      roots, weights = rysroots(n, x)

      # Compute asymptotic (x -> inf) Rys roots and weights
      # from the second half of the Hermite solution for degree 2n
      #   r_rys = r_h^2 / x
      #   w_rys = w_h / sqrt(x)
      y = one
      if (x != 0):
        y = one/x
      roots_asymp = [r*r*y for r in roots_herm[n:]]
      weights_asymp = [w*mp.sqrt(y) for w in weights_herm[n:]]

      # Compute maximum error in roots and weights, and stop when below accuracy
      error_roots = max([abs(r1-r2)/r2 for r1,r2 in zip(roots_asymp, roots)])
      error_weights = max([abs(w1-w2)/w2 for w1,w2 in zip(weights_asymp, weights)])
      error_tot = max(error_roots, error_weights)
      print('x = {0}: error = {1}'.format(figs(x, 2), figs(error_tot, 4, exp=True)))
      if (error_tot < acc):
        asymp.append(x)
        break

  # Take a step back for the next pass
  init = [a-s for a in asymp]

  print()
  print('Step: {0}'.format(mp.nstr(s)))
  for n,a in zip(order, asymp):
    print('n = {0}: x = {1}'.format(n, figs(a, 2)))
